/*  			    
 * FLAC (Free Lossless Audio Codec) decoder
 * Copyright (c) 2006 SkipJam Corp. ALL RIGHTS RESERVED.
 *  			    
 */ 			    
    			    
#include <limits.h>	    
#include <fiolib.h>
#include <stdio.h>
#include <fcntl.h>

#define FLAC_PLUGIN__MAX_SUPPORTED_CHANNELS 2
#define SAMPLES_PER_WRITE           576
#define FLAC__INLINE
#include "FLAC/all.h"

#define min(_x, _y) (((_x) < (_y)) ? (_x) : (_y))

void pstrcpy(char *buf, int buf_size, const char *str)
{
    int c;
    char *q = buf;

    if (buf_size <= 0)
        return;

    for(;;) {
        c = *str++;
        if (c == 0 || q >= buf + buf_size - 1)
            break;
        *q++ = c;
    }
    *q = '\0';
}
    			    
#define FLAC_plugin__show_error  printf
    			    
typedef struct {	    
        FIO_FUNC *fio;
        int file_opened;
        long long file_size;
        long long pending_seek;
        FLAC__StreamDecoder *decoder;
    	volatile FLAC__bool is_playing;
    	volatile FLAC__bool abort_flag;
    	volatile FLAC__bool eof;
    	volatile int seek_to;
    	unsigned total_samples;
    	unsigned bits_per_sample;
    	unsigned output_bits_per_sample;
    	unsigned channels;  
    	unsigned sample_rate;
    	unsigned length_in_msec;
	unsigned average_bps;
        FLAC__int64 decode_position;
        unsigned wide_samples_in_reservoir;
        FLAC__int32 *reservoir[FLAC_PLUGIN__MAX_SUPPORTED_CHANNELS];
        char title[256];
        char artist[256];
        char album[256];
        char genre[256];
        int track_number;
        int year;
} file_info_struct;

unsigned int g_flac_total_samples;
unsigned int g_flac_bits_per_sample;
unsigned int g_flac_sample_rate;
unsigned int g_flac_num_channels;

static FLAC__StreamDecoderReadStatus read_callback (const FLAC__StreamDecoder *decoder, FLAC__byte buffer[], unsigned *bytes, void *client_data)
{
	file_info_struct *file_info = (file_info_struct *)client_data;
	int requested_bytes = *bytes;

	if (file_info->abort_flag)
	    return FLAC__STREAM_DECODER_READ_STATUS_ABORT;

	if (file_info->eof) {
	    *bytes = 0;
	    return FLAC__STREAM_DECODER_READ_STATUS_END_OF_STREAM;
	}

	if (requested_bytes <= 0) {
	    *bytes = 0;
	    return FLAC__STREAM_DECODER_READ_STATUS_ABORT;
	}

	if (file_info->pending_seek != -1) {
	    fio_lseek(file_info->fio, file_info->pending_seek, 0);
	    file_info->pending_seek = -1;
	    FLAC__stream_decoder_flush(file_info->decoder);
	}

	*bytes = fio_read(file_info->fio, buffer, requested_bytes);
	if (*bytes == 0) {
	    file_info->eof = true;
	    return FLAC__STREAM_DECODER_READ_STATUS_END_OF_STREAM;
	}
	return FLAC__STREAM_DECODER_READ_STATUS_CONTINUE;
}

static FLAC__StreamDecoderWriteStatus write_callback(const FLAC__StreamDecoder *decoder, const FLAC__Frame *frame, const FLAC__int32 * const buffer[], void *client_data)
{
	file_info_struct *file_info = (file_info_struct*)client_data;
	const unsigned channels = file_info->channels, wide_samples = frame->header.blocksize;
	unsigned channel;

	(void)decoder;

	//printf("Write %d\n", wide_samples);
	if (file_info->abort_flag)
		return FLAC__STREAM_DECODER_WRITE_STATUS_ABORT;

	for (channel = 0; channel < channels; channel++)
		memcpy(&file_info->reservoir[channel][file_info->wide_samples_in_reservoir], buffer[channel], sizeof(buffer[0][0]) * wide_samples);

	file_info->wide_samples_in_reservoir += wide_samples;

	return FLAC__STREAM_DECODER_WRITE_STATUS_CONTINUE;
}

static char *metadata_find(const FLAC__StreamMetadata *metadata, char *name)
{
    int offset;
    const FLAC__StreamMetadata_VorbisComment_Entry *entry;
    char *p;

    offset = FLAC__metadata_object_vorbiscomment_find_entry_from(metadata, 0, name);
    if (offset < 0)
	return 0;

    entry = metadata->data.vorbis_comment.comments + offset;
    p = (char *)entry->entry;
    p = strchr(p, '=');
    if(0 == p)
	return 0;
    return ++p;
}

static void metadata_callback(const FLAC__StreamDecoder *decoder, const FLAC__StreamMetadata *metadata, void *client_data)
{
	file_info_struct *file_info = (file_info_struct*)client_data;
	char *p;
	(void)decoder;

	//printf("Metadata callback\n");
	if (metadata->type == FLAC__METADATA_TYPE_STREAMINFO)
	{
		FLAC__ASSERT(metadata->data.stream_info.total_samples < 0x100000000); /* this plugin can only handle < 4 gigasamples */
		g_flac_total_samples = file_info->total_samples = (unsigned)(metadata->data.stream_info.total_samples&0xfffffffful);
		g_flac_bits_per_sample = file_info->bits_per_sample = metadata->data.stream_info.bits_per_sample;
		g_flac_num_channels = file_info->channels = metadata->data.stream_info.channels;
		g_flac_sample_rate = file_info->sample_rate = metadata->data.stream_info.sample_rate;
		

		if (file_info->bits_per_sample!=8 && file_info->bits_per_sample!=16 && file_info->bits_per_sample!=24)
		{
			FLAC_plugin__show_error("This plugin can only handle 8/16/24-bit samples.");
			file_info->abort_flag = true;
			return;
		}
		file_info->length_in_msec = (unsigned)((double)file_info->total_samples / (double)file_info->sample_rate * 1000.0 + 0.5);
	}
	else if (metadata->type == FLAC__METADATA_TYPE_VORBIS_COMMENT)
	{
	    p = metadata_find(metadata, "TITLE");
	    if (p) {
		pstrcpy(file_info->title, sizeof(file_info->title), p);
		//fprintf(stderr, "title = %s\n", p);
	    }
	    p = metadata_find(metadata, "TRACKNUMBER");
	    if (p) {
		file_info->track_number = atoi(p);
		//fprintf(stderr, "tracknumber = %s\n", p);
	    }
	    p = metadata_find(metadata, "ALBUMARTIST");
	    if (!p)
		p = metadata_find(metadata, "ARTIST");
	    if (p) {
		pstrcpy(file_info->artist, sizeof(file_info->artist), p);
	        //fprintf(stderr, "artist = %s\n", p);
	    }
	    p = metadata_find(metadata, "ALBUM");
	    if (p) {
		pstrcpy(file_info->album, sizeof(file_info->album), p);
		//fprintf(stderr, "album = %s\n", p);
	    }
	    p = metadata_find(metadata, "GENRE");
	    if (p) {
		pstrcpy(file_info->genre, sizeof(file_info->genre), p);
		//fprintf(stderr, "genre = %s\n", p);
	    }
	    p = metadata_find(metadata, "DATE");
	    if (p) {
		file_info->year = atoi(p);
		//fprintf(stderr, "date = %s\n", p);
	    }
	}
}

static void error_callback(const FLAC__StreamDecoder *decoder, FLAC__StreamDecoderErrorStatus status, void *client_data)
{
	file_info_struct *file_info = (file_info_struct*)client_data;
	(void)decoder;

	fprintf(stderr, "flac error callback %d\n", status);
	if (status!=FLAC__STREAM_DECODER_ERROR_STATUS_LOST_SYNC)
		file_info->abort_flag = true;
}

void FLAC_plugin__decoder_finish(FLAC__StreamDecoder *decoder)
{
        if (decoder && FLAC__stream_decoder_get_state(decoder)!=FLAC__STREAM_DECODER_UNINITIALIZED)
		FLAC__stream_decoder_finish(decoder);
}

void FLAC_plugin__decoder_delete(FLAC__StreamDecoder *decoder, file_info_struct *file_info)
{
	if (decoder)
	{
		FLAC_plugin__decoder_finish(decoder);
		FLAC__stream_decoder_delete(decoder);
	}

	if (file_info->reservoir[0])
	        free(file_info->reservoir[0]);
	if (file_info->reservoir[1])
	        free(file_info->reservoir[1]);
}

FLAC__bool FLAC_plugin__decoder_init(FLAC__StreamDecoder *decoder, file_info_struct *file_info)
{
	FLAC__ASSERT(decoder);
	FLAC_plugin__decoder_finish(decoder);
	/* init decoder */
	//FLAC__stream_decoder_set_md5_checking(decoder, false);
	//FLAC__stream_decoder_set_filename(decoder, filename);
	FLAC__stream_decoder_set_metadata_ignore_all(decoder);
	FLAC__stream_decoder_set_metadata_respond(decoder, FLAC__METADATA_TYPE_STREAMINFO);
	FLAC__stream_decoder_set_metadata_respond(decoder, FLAC__METADATA_TYPE_VORBIS_COMMENT);
	FLAC__stream_decoder_set_metadata_callback(decoder, metadata_callback);
	FLAC__stream_decoder_set_read_callback (decoder, read_callback);
	FLAC__stream_decoder_set_write_callback(decoder, write_callback);
	FLAC__stream_decoder_set_error_callback(decoder, error_callback);
	FLAC__stream_decoder_set_client_data(decoder, file_info);

	//printf("Initialize decoder\n");
	if (FLAC__stream_decoder_init(decoder) != FLAC__STREAM_DECODER_SEARCH_FOR_METADATA)
	{
		FLAC_plugin__show_error("Error while initializing decoder (%d).", FLAC__stream_decoder_get_state(decoder));
		return false;
	}
	/* process */
	file_info->wide_samples_in_reservoir = 0;
	file_info->is_playing = false;
	file_info->abort_flag = false;

	//printf("Process metadata\n");
	if (!FLAC__stream_decoder_process_until_end_of_metadata(decoder))
	{
		FLAC_plugin__show_error("Error while processing metadata (%d).", FLAC__stream_decoder_get_state(decoder));
		return false;
	}
	/* check results */
	if (file_info->abort_flag) return false;                /* metadata callback already popped up the error dialog */
	/* init replaygain */
	file_info->output_bits_per_sample = 16;

	/* more inits */
	file_info->eof = false;
	file_info->seek_to = -1;
	file_info->is_playing = true;
	//file_info->average_bps = (unsigned)(filesize / (125.*file_info->total_samples/file_info->sample_rate));
	
	file_info->decode_position = 0;
	//written_time_last = 0;

	file_info->reservoir[0] = malloc(FLAC__MAX_BLOCK_SIZE * 2);
	file_info->reservoir[1] = malloc(FLAC__MAX_BLOCK_SIZE * 2);

	//printf("Init done\n");
	return true;
}

static FLAC__INLINE FLAC__uint32 prng(FLAC__uint32 state)
{
	return (state * 0x0019660dL + 0x3c6ef35fL) & 0xffffffffL;
}

/* dither routine derived from MAD winamp plugin */

typedef struct {
	FLAC__int32 error[3];
	FLAC__int32 random;
} dither_state;

static FLAC__INLINE FLAC__int32 linear_dither(unsigned source_bps, unsigned target_bps, FLAC__int32 sample, dither_state *dither, const FLAC__int32 MIN, const FLAC__int32 MAX)
{
	unsigned scalebits;
	FLAC__int32 output, mask, random;

	FLAC__ASSERT(source_bps < 32);
	FLAC__ASSERT(target_bps <= 24);
	FLAC__ASSERT(target_bps <= source_bps);

	/* noise shape */
	sample += dither->error[0] - dither->error[1] + dither->error[2];

	dither->error[2] = dither->error[1];
	dither->error[1] = dither->error[0] / 2;

	/* bias */
	output = sample + (1L << (source_bps - target_bps - 1));

	scalebits = source_bps - target_bps;
	mask = (1L << scalebits) - 1;

	/* dither */
	random = (FLAC__int32)prng(dither->random);
	output += (random & mask) - (dither->random & mask);

	dither->random = random;

	/* clip */
	if(output > MAX) {
		output = MAX;

		if(sample > MAX)
			sample = MAX;
	}
	else if(output < MIN) {
		output = MIN;

		if(sample < MIN)
			sample = MIN;
	}

	/* quantize */
	output &= ~mask;

	/* error feedback */
	dither->error[0] = sample - output;

	/* scale */
	return output >> scalebits;
}

size_t FLAC__plugin_common__pack_pcm_signed_little_endian(FLAC__byte *data, const FLAC__int32 * const input[], unsigned wide_samples, unsigned channels, unsigned source_bps, unsigned target_bps)
{
	static dither_state dither[FLAC_PLUGIN__MAX_SUPPORTED_CHANNELS];
	FLAC__byte * const start = data;
	FLAC__int32 sample;
	const FLAC__int32 *input_;
	unsigned samples, channel;
	const unsigned bytes_per_sample = target_bps / 8;
	const unsigned incr = bytes_per_sample * channels;

	FLAC__ASSERT(channels > 0 && channels <= FLAC_PLUGIN__MAX_SUPPORTED_CHANNELS);
	FLAC__ASSERT(source_bps < 32);
	FLAC__ASSERT(target_bps <= 24);
	FLAC__ASSERT(target_bps <= source_bps);
	FLAC__ASSERT((source_bps & 7) == 0);
	FLAC__ASSERT((target_bps & 7) == 0);

	if(source_bps != target_bps) {
		const FLAC__int32 MIN = -(1L << (source_bps - 1));
		const FLAC__int32 MAX = ~MIN; /*(1L << (source_bps-1)) - 1 */

		for(channel = 0; channel < channels; channel++) {
			
			samples = wide_samples;
			data = start + bytes_per_sample * channel;
			input_ = input[channel];

			while(samples--) {
				sample = linear_dither(source_bps, target_bps, *input_++, &dither[channel], MIN, MAX);

				switch(target_bps) {
					case 8:
						data[0] = sample ^ 0x80;
						break;
					case 24:
						data[2] = (FLAC__byte)(sample >> 16);
						/* fall through */
					case 16:
						data[1] = (FLAC__byte)(sample >> 8);
						data[0] = (FLAC__byte)sample;
				}

				data += incr;
			}
		}
	}
	else {
		for(channel = 0; channel < channels; channel++) {
			samples = wide_samples;
			data = start + bytes_per_sample * channel;
			input_ = input[channel];

			while(samples--) {
				sample = *input_++;

				switch(target_bps) {
					case 8:
						data[0] = sample ^ 0x80;
						break;
					case 24:
						data[2] = (FLAC__byte)(sample >> 16);
						/* fall through */
					case 16:
						data[1] = (FLAC__byte)(sample >> 8);
						data[0] = (FLAC__byte)sample;
				}

				data += incr;
			}
		}
	}

	return wide_samples * channels * (target_bps/8);
}

unsigned FLAC_plugin__decode(FLAC__StreamDecoder *decoder, file_info_struct *file_info, char *sample_buffer, int num_samples)
{
	/* fill reservoir */
        while (file_info->wide_samples_in_reservoir < num_samples)
	{
		if (FLAC__stream_decoder_get_state(decoder) == FLAC__STREAM_DECODER_END_OF_STREAM)
		{
			file_info->eof = true;
			break;
		}
		else if (!FLAC__stream_decoder_process_single(decoder))
		{
			FLAC_plugin__show_error("Error while processing frame (%d).", FLAC__stream_decoder_get_state(decoder));
			file_info->eof = true;
			break;
		}
		//if (!FLAC__stream_decoder_get_decode_position(decoder, &file_info->decode_position))
		//file_info->decode_position = 0;
	}
	/* output samples */
	if (file_info->wide_samples_in_reservoir > 0)
	{
	        const unsigned n = min(file_info->wide_samples_in_reservoir, num_samples);
		const unsigned channels = file_info->channels;
		unsigned i;
		int bytes;

		bytes = FLAC__plugin_common__pack_pcm_signed_little_endian(
		    sample_buffer,
		    &file_info->reservoir,
		    n,
		    channels,
		    file_info->bits_per_sample,
		    file_info->output_bits_per_sample
		    );

		file_info->wide_samples_in_reservoir -= n;
		for (i = 0; i < channels; i++)
			memmove(&file_info->reservoir[i][0], &file_info->reservoir[i][n],
				sizeof(file_info->reservoir[0][0]) * file_info->wide_samples_in_reservoir);

		return bytes;
	}
	else
	{
		file_info->eof = true;
		return 0;
	}
}

int FLAC_Open(FIO_FUNC *fio, void **ppvHandle)
{
    file_info_struct *file_info;

    file_info = malloc(sizeof(file_info_struct));
    if (!file_info)
	return -1;

    memset(file_info, 0, sizeof(file_info_struct));
    file_info->fio = fio;
    file_info->pending_seek = -1;

    file_info->file_size = fio_lseek(fio, 0, SEEK_END);
    fio_lseek(fio, 0, SEEK_SET);

    file_info->decoder = FLAC__stream_decoder_new();

    if (!FLAC_plugin__decoder_init(file_info->decoder, file_info)) {
	FLAC_plugin__decoder_delete(file_info->decoder, file_info);
	free(file_info);
	return -1;
    }

    *ppvHandle = file_info;
    return 0;
}

int FLAC_OpenFile(char *FileName, void **ppvHandle)
{
    file_info_struct *file_info;
    FIO_FUNC *fio;

    fio = malloc(sizeof(FIO_FUNC));
    if (!fio)
	return -1;

    if (fio_open(FileName, O_RDONLY, 0, fio) < 0) {
	free(fio);
	return -1;
    }

    if (FLAC_Open(fio, &file_info) < 0) {
	fio_close(fio);
	free(fio);
	return -1;
    }

    file_info->file_opened = 1;
    *ppvHandle = file_info;
    return 0;
}

int FLAC_Read(void *pvHandle, unsigned char *buffer, int length)
{
    file_info_struct *file_info = pvHandle;
    int num_samples;

    num_samples = length / (FLAC_PLUGIN__MAX_SUPPORTED_CHANNELS * (24/8) * 2);
    return FLAC_plugin__decode(file_info->decoder, file_info, buffer, num_samples);
}

int FLAC_Seek(void *pvHandle, int whence, long long pos)
{
    file_info_struct *file_info = pvHandle;
    file_info->pending_seek = pos;
    return 0;
}

int FLAC_GetFileSize(void *pvHandle, long long *file_size)
{
    file_info_struct *file_info = pvHandle;
    *file_size = file_info->file_size;
    return 0;
}

int FLAC_Close(void *pvHandle)
{
    file_info_struct *file_info = pvHandle;

    FLAC_plugin__decoder_delete(file_info->decoder, file_info);
    if (file_info->file_opened)
	fio_close(file_info->fio);
    free(file_info);
    return 0;
}

int FLAC_GetStreamInfo(void *pvHandle, unsigned int *total_samples, unsigned int *bits_per_sample, unsigned int *num_channels, unsigned int *sample_rate)
{
    file_info_struct *file_info = pvHandle;

    if (total_samples)
	*total_samples = file_info->total_samples;
    if (bits_per_sample)
	*bits_per_sample = file_info->bits_per_sample;
    if (num_channels)
	*num_channels = file_info->channels;
    if (sample_rate)
	*sample_rate = file_info->sample_rate;
    return 0;
}

int FLAC_GetDuration(void *pvHandle, long long *msecs)
{
    file_info_struct *file_info = pvHandle;
    *msecs = file_info->length_in_msec;
    return 0;
}

int FLAC_GetTagInfo(void *pvHandle, char *title, int title_len, int *track_number, char *artist, int artist_len, char *album, int album_len,
		    char *genre, int genre_len, int *year)
{
    file_info_struct *file_info = pvHandle;

    if (title)
	pstrcpy(title, title_len, file_info->title);
    if (track_number)
	*track_number = file_info->track_number;
    if (artist)
	pstrcpy(artist, artist_len, file_info->artist);
    if (album)
	pstrcpy(album, album_len, file_info->album);
    if (genre)
	pstrcpy(genre, genre_len, file_info->genre);
    if (year)
	*year = file_info->year;
    return 0;
}

