
//-------------------------------------------------------------------------
// INTERRUPTIBLE.C - Interruptible section helper
//
// Copyright (c) 2008 NETGEAR, Inc.
// All rights reserved.
//-------------------------------------------------------------------------

#include <stdio.h>
#include <pthread.h>
#include <signal.h>

#include "interruptible.h"

static pthread_mutex_t 		g_interruptible_mutex;
static pthread_key_t		g_interruptible_key;

//
// interruptible_dispatcher - dispatches signal to section handlers
//
// entry: pintr = supplies the interruptible section
// exit:  returns nonzero if the signal was/should be handled
//
int interruptible_dispatcher(interruptible_t *pintr, int sig)
{
    interruptible_t *pnext;
    int pid = getpid();

    printf("[%d] interruptible_dispatcher, pintr = %x, sig = %d\n", pid, pintr, sig);

    pnext = pintr->pnext;
    if (pnext) {

	if (!interruptible_dispatcher(pnext, sig))
	    return 0;

	pintr->signalled = pnext->signalled;
    }

    switch ((int)pintr->handler) {
    case INTERRUPTIBLE_NEVER:
	printf("[%d] interruptible_dispatcher: Interrupts are disabled\n", pid);
	pintr->signalled = 0;
	return 0;
    case INTERRUPTIBLE_ALWAYS:
	printf("[%d] interruptible_dispatcher: Interrupts are enabled\n", pid);
	pintr->signalled = sig;
	return 1;
    case INTERRUPTIBLE_DEFAULT:
	printf("[%d] interruptible_dispatcher: Default action\n", pid);
	return 1;
    case INTERRUPTIBLE_ALARM:
	if (pintr->signalled && sig != SIGALRM) {
	    printf("[%d] interruptible_dispatcher: Raising SIGALRM\n", pid);
	    raise(SIGALRM);
	}
	return 1;
    default:
	printf("[%d] interruptible_dispatcher: Calling handler at %x\n", pid, pintr->handler);
	return pintr->handler(pintr);
    }
}

//
// interruptible_sighandler - internal signal dispatcher
//
// entry: sig = signal id
// exit:  none
//
void interruptible_sighandler(int sig)
{
    interruptible_t *pintr;
    int pid = getpid();

    printf("[%d] interruptible_sighandler, sig %d\n", pid, sig);
    
    if (sig != INTERRUPTIBLE_SIG && sig != SIGALRM) {
	printf("[%d] interruptible_sighandler: Unexpected signal %d\n", pid, sig);
	return;
    }

    pintr = (interruptible_t *)pthread_getspecific(g_interruptible_key);
    if (!pintr) {
	printf("[%d] interruptible_sighandler: Not in an interruptible section\n", pid);
	return;
    }

    interruptible_dispatcher(pintr, sig);
}

//
// init_interruptible - initialize globals for this helper module
//
// entry: none
// exit:  returns nonzero
//
int init_interruptible(void)
{
    struct sigaction newact;

    pthread_mutex_init(&g_interruptible_mutex, NULL);
    pthread_key_create(&g_interruptible_key, 0);

    newact.sa_handler = interruptible_sighandler;
    sigemptyset(&newact.sa_mask);
    newact.sa_flags = SA_RESTART;
    sigaction(INTERRUPTIBLE_SIG, &newact, NULL);

    newact.sa_handler = interruptible_sighandler;
    sigemptyset(&newact.sa_mask);
    newact.sa_flags = 0;
    sigaction(SIGALRM, &newact, NULL);

    return 1;
}

//
// is_intrerrupted - tests whether interrupt is pending for current pid
//
// entry: none
// exit:  returns nonzero if interrupt is pending
//
int is_interrupted(void)
{
    interruptible_t *pintr;

    pintr = (interruptible_t *)pthread_getspecific(g_interruptible_key);
    if (!pintr)
	return 0;
    return pintr->signalled;
}

//
// enter_interruptible - enter an interruptible section
//
// entry: none
// exit:  returns zero if pid was already interrupted
//
int enter_interruptible(interruptible_t *pintr)
{
    interruptible_t *pnext;

#if 0
    unsigned long ipaddr;

    __asm__ __volatile__ (
	"mov	%0, lr\n"
	:"=r" (ipaddr)
	);
#endif

    //printf("[%d] enter_interruptible, pintr %x, handler = %x, ipaddr = %x\n", getpid(), pintr, pintr->handler, ipaddr);
    pnext = (interruptible_t *)pthread_getspecific(g_interruptible_key);
    if (pnext) {

	if (pnext->signalled) {
	    printf("[%d] enter_interruptible: already signalled\n", getpid());
	    pintr->signalled = pnext->signalled;
	    return 0;
	}

	pintr->pnext = pnext;
    }

    pintr->pnext = pnext;
    pintr->signalled = 0;
    pthread_setspecific(g_interruptible_key, pintr);
    return 1;
}

//
// leave_interruptible - leave an interruptible section
//
// entry: none
// exit:  none
//
void leave_interruptible(interruptible_t *pintr)
{
    interruptible_t *pnext;

    //printf("[%d] leave_interruptible\n", getpid());
    pnext = pintr->pnext;
    if (pnext && pintr->signalled)
	pnext->signalled = pintr->signalled;
    pthread_setspecific(g_interruptible_key, pnext);
}

//
// reset_interruptible - resets the interrupt status of the current pid
//
// entry: none
// exit:  none
//
void reset_interruptible(void)
{
    interruptible_t *pnext;

    //printf("[%d] reset_interruptible\n", getpid());

    pnext = (interruptible_t *)pthread_getspecific(g_interruptible_key);
    while (pnext) {
	pnext->signalled = 0;
	pnext = pnext->pnext;
    }
}

//
// interrupt_pid - interrupts a pid
//
// entry: none
// exit:  none
//
void interrupt_pid(int pid)
{
    printf("[%d] interrupt_pid: Signalling pid %d\n", getpid(), pid);
    kill(pid, INTERRUPTIBLE_SIG);
}

//
// interrupt_section - interrupts a section
//
// entry: pintr = supplies the section
// exit:  non
//
int interrupt_section(interruptible_t *pintr)
{
    printf("[%d] interrupt_section, pintr = %x\n", getpid(), pintr);
    if (!pintr->pid)
	return 0;
    interrupt_pid(pintr->pid);
    return 1;
}
	



