/**
 * Filename: file.c
 * 
 * Created by : Dennis J. Aldover (daldover@skipjam.com)
 * 
 * Copyright 2005, SkipJam Inc. All rights reserved
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <time.h>
#include <assert.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include "libsmbclient.h"
#ifdef _SJMENUS
#include "MemoryManagement.h"
#include "contract.h"
#include "utility.h"
#include "errorcodes.h"
#endif //_SJMENUS
#include "file.h"

/**
 * 
 */
const char* STR_PROTOCOL_SMB= "smb://";

/**
 * 
 */
void exif_smb_auth_fn(const char *server, const char *share, char *workgroup, 
					  int wgmaxlen, char *username, int unmaxlen, char *password, int pwmaxlen);
char *exif_smbc_fgets(char *str, int n, int fh);

/**
 * 
 */
const char* ExifFile_GetProtocol( PtrFile f )
{
	if( f && f->fn && strlen(f->fn) ) {
	  //fprintf(stderr,"libexif: [ExifFile_GetProtocol] - '%s'\n", f->fn);
		if( !strncasecmp(f->fn,STR_PROTOCOL_SMB,strlen(STR_PROTOCOL_SMB))){
		  //fprintf(stderr,"libexif: [ExifFile_GetProtocol] - PROTOCOL_SMB\n" );
			return STR_PROTOCOL_SMB;
		} else {
		  //fprintf(stderr,"libexif: [ExifFile_GetProtocol] - Protocol not supported\n" );
		}
	}
	return 0;
}

/**
 * 
 */
PtrFile ExifFile_( const char* path )
{
	PtrFile p = 0;
	int err= 0;
	int len= 0;


	IS_VALID_PTR(path,err);
	//fprintf(stderr,"ExifFile_(%s)\n",path);
	TRY_ALLOC(p,1,sizeof(File),"ExifFile_",err);
	TRY_ALLOC(p->fn,(len=strlen(path))+1,sizeof(char),"ExifFile_",err);
	strcpy(p->fn,path);
	p->fn[len]= 0;
	p->h= 0;

SAFE_EXIT_HANDLER:
	return p;
}

/**
 * 
 */
void ExifFile__( PtrFile p )
{
	int err= 0;

	IS_VALID_PTR(p,err);
	//fprintf(stderr,"ExifFile__(%s)\n",p->fn);
	if(ExifFile_IsOpen(p)) {
		ExifFile_Close(p);
	}
	TRY_FREE(p->fn,err);
	TRY_FREE(p,err);
SAFE_EXIT_HANDLER:
	return;
}

void exif_smb_auth_fn(const char *server, const char *share, char *workgroup,
			int wgmaxlen, char *username, int unmaxlen, char *password, int pwmaxlen)
{
    username[0] = 0;
    password[0] = 0;
}

/**
 * 
 */
int ExifFile_Open( PtrFile p, int flags, mode_t mode )
{
  //fprintf(stderr, ">>> ExifFile_Open\n");
	if( p ) {
	  if( strlen( p->fn )) {
	    //fprintf(stderr,"ExifFile_Open: filename: '%s'\n", p->fn);
	    if( ExifFile_GetProtocol(p) == STR_PROTOCOL_SMB ) {
	      //fprintf(stderr,"ExifFile_Open: smbc_open('%s',%d,%d)\n",p->fn,flags,mode );
				smbc_init(exif_smb_auth_fn,0);
	      p->h = smbc_open( p->fn, flags, mode );	      
	    }
	    else  {
	      //fprintf(stderr,"ExifFile_Open: open('%s',%d,%d)\n",p->fn,flags,mode );
	      p->h = open( p->fn, flags, mode );
	    }
	    return p->h;
	  }
	}
	//fprintf(stderr, "<<< ExifFile_Open\n");
	return 0;
}

/**
 * 
 */
int ExifFile_IsOpen( PtrFile p )
{
	if( p ) {
		return p->h;
	}
	return 0;
}

/**
 * 
 */
void ExifFile_Close( PtrFile p )
{
	if( p && (p->h>0) ) {
		if( ExifFile_GetProtocol(p) == STR_PROTOCOL_SMB ) {
			smbc_close( p->h );
		}
		else {
			close( p->h );
		}
		p->h= 0;
	}
}

/**
 * 
 */
int ExifFile_Handle( PtrFile p )
{
	if( p ) {
		return p->h;
	}
	return -1;
}

/**
 * 
 */
ssize_t	ExifFile_Read( PtrFile p, void* buf, size_t count )
{
	if( p ) {
		if( ExifFile_GetProtocol(p) == STR_PROTOCOL_SMB ) {
			return smbc_read(p->h,buf,count);
		}
		else {
			return read(p->h,buf,count);
		}
	}
	return -1;
}

/**
 * 
 */
ssize_t ExifFile_Write( File * p, const char * buf, size_t count )
{
	if( p ) {
		if( ExifFile_GetProtocol(p) == STR_PROTOCOL_SMB ) {
			return smbc_write( p->h,buf,count );
		}
		else  {			
			//DBG_PRINT(0,"ExifFile_Write: '%s' count(%d)\n", buf, count);
			return write( p->h, buf, count );
		}
	}

	return -1;
}

/**
 * 
 */
char *exif_smbc_fgets(char *str, int n, int fh)
{
    static char buf[EXIF_FIO_BUF_SIZE];
    static char *next;
	static int r;
    char *s;
    int c;

    c = 0;
    s = str;
    while( c < n ) {
        if((next == 0) || (next == (buf+r))) {
            r = smbc_read(fh, buf, EXIF_FIO_BUF_SIZE );
            if(r < 0)                             //error
                return NULL;
            if(! r ) {                            //end of file
                *s = 0;
                next = 0;
                if(c)                             //if we read any bytes
                    return str;
                else
                    return NULL;
            }
            next = buf;
        }
        if(*next < 32) {
            if(*next == '\n') {
                *s = 0;
                ++next;
                break;
            }
            ++next;
            continue;
        }

        *s++ = *next++;
        ++c;
    }
    return str;
}


/**
 * 
 */
char* ExifFile_FGets( PtrFile p, char* buf, int n )
{
	//DBG_PRINT(0,">>> ExifFile_FGets - File: '%s'\n\tbuffer size: %d\n\thandle: %d\n", p->fn, n,p->h);
	if( p ) {
		if( ExifFile_GetProtocol(p) == STR_PROTOCOL_SMB ) {
			return (char*)exif_smbc_fgets( buf, n, p->h );
		}
		else {
			int i=0;
			while((i<n)) {
				
				if(0>=ExifFile_Read(p,&buf[i],1)) {
					break;
				}
				if((buf[i]=='\n')){
					i++;
					break;
				}
				i++;
			}
			buf[i]=0;
			return (buf[0]!=0)?buf:0;
		}
	}
	return 0;
}

/**
 * 
 */
long long ExifFile_LSeek( PtrFile p, off_t offset, int mode )
{
	long long sk= 0;
	int err= 0;
	int close= 0;

	IS_VALID_PTR(p,err);
	if(!ExifFile_IsOpen(p)){
		if(ExifFile_Open(p,O_RDONLY,S_IREAD)>0){
			close= 1;
		}
	}
	if( ExifFile_IsOpen(p)) {
		if( ExifFile_GetProtocol(p) == STR_PROTOCOL_SMB ) {
			DBG_PRINT(0,"ExifFile_LSeek: smbc_lseek\n");
			sk = smbc_lseek(p->h, (long long)offset, mode );
		} else {
			sk = lseek(p->h,offset,mode);			
		}
		if(sk<0) {
			err= errno;
		}
	} else {
		DBG_PRINT(0,"ExifFile_LSeek: file is not opened!\n");
	}
SAFE_EXIT_HANDLER:
	if(close){
		ExifFile_Close(p);
	}
	return sk;
}

/**
 * 
 */
long long ExifFile_GetSize( PtrFile p )
{
  //struct stat st={0};
	int 		err=0;
	long long	size =0;

	IS_VALID_PTR(p,err);

	size= ExifFile_LSeek(p,0,SEEK_END);

SAFE_EXIT_HANDLER:
	if( err != 0 ) {
		fprintf( stderr, "ExifFile_GetSize: ERR(%d)\n", err );
	} else {
		DBG_PRINT(0,"ExifFile_GetSize: %lld\n", size);
	}

	return size;
}
