
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <malloc.h>
#include <errno.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <libsmbclient.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <pthread.h>
#include "interruptible.h"

#include "smbc_api.h"

#define smbc_enter_mutex()	pthread_mutex_lock(&smbc_mutex)
#define smbc_leave_mutex()	pthread_mutex_unlock(&smbc_mutex)
#define FIELD_OFFSET(type, field)  ((long)&(((type *)0)->field))
#define arrayof(x) 		(sizeof(x) / sizeof(*(x)))
#define min(x,y)		((x < y) ? x : y)
char *smbc_strcpyn(char *dst, char *src, int len);

static int smbc_fd = -1;
static int smbc_initialized = 0;
static int smbc_client_pid = 0;
static pthread_mutex_t smbc_mutex;
static smbc_get_auth_data_fn smbc_auth_fn = NULL;

//#define SMBC_DEBUG(msg)		printf(msg)
#define SMBC_DEBUG(msg)

#ifdef SMBC_USE_SHARED_MEMORY

#define SMBC_SHARED_BUF_SIZE (128*1024)
static int smbc_shared_h = -1;
static void *smbc_shared_buf;

static struct {
    unsigned char *base_address;
    int		   size;
    int		   id;
} g_smbc_shmaps[10];
#endif

int smbc_init(smbc_get_auth_data_fn fn, int debug)
{
    struct sockaddr_in  addr;
    pthread_mutexattr_t attr;
    SMBC_ACK_INFO       ack_info;

    if (!smbc_initialized) {
	pthread_mutexattr_init(&attr);
	pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE_NP);
	pthread_mutex_init(&smbc_mutex, &attr);
	smbc_initialized = 1;
    }

    smbc_enter_mutex();
    if (smbc_fd < 0) {
	memset(&addr, 0, sizeof(addr));
	addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	addr.sin_port        = htons(SMBC_REMOTE_PORT);
	addr.sin_family      = AF_INET;
	smbc_fd = socket(AF_INET, SOCK_STREAM, 0);
	if (smbc_fd >= 0 && connect(smbc_fd, &addr, sizeof(addr)) < 0) {
	    close(smbc_fd);
	    smbc_fd = -1;
	}

	ack_info.hdr.size = sizeof(SMBC_MSG);
	ack_info.hdr.type = SMBC_MSG_GETPID;
	if (!smbc_send_recv_msg(&ack_info, &ack_info)) {
	    printf("smbc_init: Couldn't query smbclient pid\n");
	}
	else {
	    smbc_client_pid = ack_info.value.iValue;
	    printf("smbc_init: Smbclient pid is %d\n", smbc_client_pid);
	}
    }

#ifdef SMBC_USE_SHARED_MEMORY
    if (smbc_shared_h < 0) {
	smbc_shared_h = smbc_alloc_shared(SMBC_SHARED_BUF_SIZE, &smbc_shared_buf);
	if (smbc_shared_h < 0) {
	    printf("smbc_init: Failed to allocated shared handle\n");
	    close(smbc_fd);
	    smbc_fd = -1;;
	}
    }
#endif

    smbc_leave_mutex();
    
    return (smbc_fd >= 0);
}

int smbc_reinit(void)
{
    int ret;
    SMBC_MSG msg;
    SMBC_ACK_INFO ack_info;

    smbc_enter_mutex();
    msg.type = SMBC_MSG_REINIT;
    msg.size = sizeof(SMBC_MSG);
    ret = smbc_send_recv_msg(&msg, &ack_info) ? ack_info.value.iValue : -1;
    smbc_leave_mutex();
    return ret;
}

int smbc_open(const char *fname, int flags, mode_t mode)
{
    int ret;
    int i = 0;
    unsigned char *sbuf;
    SMBC_OPEN_INFO open_info;
    SMBC_ACK_INFO ack_info;

    SMBC_DEBUG(("smbc_open\n"));

    smbc_enter_mutex();
    open_info.fname    = fname;
    open_info.flags    = flags;
    open_info.mode     = mode;
    open_info.hdr.type = SMBC_MSG_OPEN;
    open_info.hdr.size = sizeof(SMBC_OPEN_INFO);
#ifdef SMBC_USE_SHARED_MEMORY
    open_info.h        = smbc_shared_h;
    open_info.fname    = smbc_wire_string(open_info.h, &i, open_info.fname);
#endif
    ret = smbc_send_recv_msg(&open_info, &ack_info) ? ack_info.value.iValue : -1;
    smbc_leave_mutex();
    return ret;
}

int smbc_creat(const char *furl, mode_t mode)
{
    return smbc_open(furl, O_CREAT|O_WRONLY|O_TRUNC, mode);
}

ssize_t smbc_read(int fd, void *buf, size_t count)
{
    int ret = 0;
    int len;
    SMBC_READ_INFO read_info;
    SMBC_ACK_INFO ack_info;

    SMBC_DEBUG(("smbc_read\n"));

    smbc_enter_mutex();

#ifdef SMBC_USE_SHARED_MEMORY
    while (count > 0) {
	len = min(count, SMBC_SHARED_BUF_SIZE);
	len = smbc_read_shared(fd, smbc_shared_buf, len, smbc_shared_h);
	if (len <= 0)
	    break;
	memcpy(buf, smbc_shared_buf, len);
	buf += len;
	ret += len;
	count -= len;
    }
#else
    read_info.fd       = fd;
    read_info.count    = count;
    read_info.buf      = buf;
    read_info.hdr.type = SMBC_MSG_READ;
    read_info.hdr.size = sizeof(SMBC_READ_INFO);
    ret = smbc_send_recv_msg(&read_info, &ack_info) ? ack_info.value.iValue : -1;
#endif
    smbc_leave_mutex();
    return ret;
}

ssize_t smbc_write(int fd, void *buf, size_t count)
{
    int ret = 0;
    int len;
    SMBC_WRITE_INFO write_info;
    SMBC_ACK_INFO ack_info;

    smbc_enter_mutex();

#ifdef SMBC_USE_SHARED_MEMORY
    while (count > 0) {
	len = min(count, SMBC_SHARED_BUF_SIZE);
	memcpy(smbc_shared_buf, buf, len);
	len = smbc_write_shared(fd, smbc_shared_buf, len, smbc_shared_h);
	if (len <= 0)
	    break;
	buf += len;
	ret += len;
	count -= len;
    }
#else
    write_info.fd       = fd;
    write_info.buf      = buf;
    write_info.count    = count;
    write_info.hdr.type = SMBC_MSG_WRITE;
    write_info.hdr.size = sizeof(SMBC_WRITE_INFO);
    ret = smbc_send_recv_msg(&write_info, &ack_info) ? ack_info.value.iValue : -1;
#endif
    smbc_leave_mutex();
    return ret;
}

long long smbc_lseek(int fd, long long offset, int whence)
{
    long long ret;
    SMBC_LSEEK_INFO lseek_info;
    SMBC_ACK_INFO ack_info;

    SMBC_DEBUG(("smbc_lseek\n"));

    smbc_enter_mutex();
    lseek_info.fd       = fd;
    lseek_info.offset   = offset;
    lseek_info.whence   = whence;
    lseek_info.hdr.type = SMBC_MSG_LSEEK;
    lseek_info.hdr.size = sizeof(SMBC_LSEEK_INFO);
    ret = smbc_send_recv_msg(&lseek_info, &ack_info) ? ack_info.value.llValue : -1;
    smbc_leave_mutex();
    return ret;
}

int smbc_close(int fd)
{
    int ret;
    SMBC_CLOSE_INFO close_info;
    SMBC_ACK_INFO ack_info;

    SMBC_DEBUG(("smbc_close\n"));

    smbc_enter_mutex();
    close_info.fd       = fd;
    close_info.hdr.type = SMBC_MSG_CLOSE;
    close_info.hdr.size = sizeof(SMBC_CLOSE_INFO);
    ret = smbc_send_recv_msg(&close_info, &ack_info) ? ack_info.value.iValue : -1;
    smbc_leave_mutex();
    return ret;
}

int smbc_unlink(const char *furl)
{
    int ret;
    int i = 0;
    SMBC_UNLINK_INFO unlink_info;
    SMBC_ACK_INFO ack_info;

    smbc_enter_mutex();
    unlink_info.furl     = furl;
    unlink_info.hdr.type = SMBC_MSG_UNLINK;
    unlink_info.hdr.size = sizeof(SMBC_UNLINK_INFO);
#ifdef SMBC_USE_SHARED_MEMORY
    unlink_info.h        = smbc_shared_h;
    unlink_info.furl     = smbc_wire_string(unlink_info.h, &i, unlink_info.furl);
#endif
    ret = smbc_send_recv_msg(&unlink_info, &ack_info) ? ack_info.value.iValue : -1;
    smbc_leave_mutex();
    return ret;
}

int smbc_lock(int fd, unsigned long offset, unsigned long len, int timeout, int lock_type)
{
    int ret;
    SMBC_LOCK_INFO lock_info;
    SMBC_ACK_INFO ack_info;

    smbc_enter_mutex();
    lock_info.fd        = fd;
    lock_info.offset    = offset;
    lock_info.len       = len;
    lock_info.timeout   = timeout;
    lock_info.lock_type = lock_type;
    lock_info.hdr.type  = SMBC_MSG_LOCK;
    lock_info.hdr.size  = sizeof(SMBC_LOCK_INFO);
    ret = smbc_send_recv_msg(&lock_info, &ack_info) ? ack_info.value.iValue : -1;
    smbc_leave_mutex();
    return ret;
}

int smbc_unlock(int fd, unsigned long offset, unsigned long len)
{
    int ret;
    SMBC_UNLOCK_INFO unlock_info;
    SMBC_ACK_INFO ack_info;

    smbc_enter_mutex();
    unlock_info.fd        = fd;
    unlock_info.offset    = offset;
    unlock_info.len       = len;
    unlock_info.hdr.type  = SMBC_MSG_UNLOCK;
    unlock_info.hdr.size  = sizeof(SMBC_UNLOCK_INFO);
    ret = smbc_send_recv_msg(&unlock_info, &ack_info) ? ack_info.value.iValue : -1;
    smbc_leave_mutex();
    return ret;
}

int smbc_rename(const char *ourl, const char *nurl)
{
    int ret;
    int i = 0;
    SMBC_RENAME_INFO rename_info;
    SMBC_ACK_INFO ack_info;

    smbc_enter_mutex();
    rename_info.ourl     = ourl;
    rename_info.nurl     = nurl;
    rename_info.hdr.type = SMBC_MSG_RENAME;
    rename_info.hdr.size = sizeof(SMBC_RENAME_INFO);
#ifdef SMBC_USE_SHARED_MEMORY
    rename_info.h        = smbc_shared_h;
    rename_info.ourl     = smbc_wire_string(rename_info.h, &i, rename_info.ourl);
    rename_info.nurl     = smbc_wire_string(rename_info.h, &i, rename_info.nurl);
#endif
    ret = smbc_send_recv_msg(&rename_info, &ack_info) ? ack_info.value.iValue : -1;
    smbc_leave_mutex();
    return ret;
}

typedef struct {
    int		sig;
    int 	fd;
    char	dirent[FIELD_OFFSET(struct smbc_dirent, name[512])];
} SMBC_DIR_HANDLE;

int smbc_opendir(const char *fname)
{
    int ret;
    int i = 0;
    SMBC_OPENDIR_INFO opendir_info;
    SMBC_ACK_INFO ack_info;
    SMBC_DIR_HANDLE *dir;

    SMBC_DEBUG(("smbc_opendir\n"));

#ifdef SMBC_USE_SHARED_MEMORY
    dir = malloc(sizeof(SMBC_DIR_HANDLE));
    if (!dir)
	return -1;
#endif
    smbc_enter_mutex();
    opendir_info.fname    = fname;
    opendir_info.hdr.type = SMBC_MSG_OPENDIR;
    opendir_info.hdr.size = sizeof(SMBC_OPENDIR_INFO);
#ifdef SMBC_USE_SHARED_MEMORY
    opendir_info.h        = smbc_shared_h;
    opendir_info.fname    = smbc_wire_string(opendir_info.h, &i, opendir_info.fname);
#endif
    ret = smbc_send_recv_msg(&opendir_info, &ack_info) ? ack_info.value.iValue : -1;
#ifdef SMBC_USE_SHARED_MEMORY
    if (ret >= 0) {
	dir->sig = 'SMBD';
	dir->fd = ret;
	ret = (int)dir;
    }
    else free(dir);
#endif
    smbc_leave_mutex();
    return ret;
}

int smbc_closedir(int fd)
{
    int ret;
    SMBC_CLOSEDIR_INFO closedir_info;
    SMBC_ACK_INFO ack_info;
    SMBC_DIR_HANDLE *dir;

#ifdef SMBC_USE_SHARED_MEMORY
    dir = (SMBC_DIR_HANDLE *)fd;
    if (dir->sig != 'SMBD') {
	printf("smbc_closedir: Invalid dir handle %d\n", fd);
	return -1;
    }
    fd = dir->fd;
    dir->sig = 0;
    free(dir);
#endif
    smbc_enter_mutex();
    closedir_info.fd      = fd;
    closedir_info.hdr.type = SMBC_MSG_CLOSEDIR;
    closedir_info.hdr.size = sizeof(SMBC_CLOSEDIR_INFO);
    ret = smbc_send_recv_msg(&closedir_info, &ack_info) ? ack_info.value.iValue : -1;
    smbc_leave_mutex();
    return ret;
}

struct smbc_dirent *smbc_readdir(unsigned int fd)
{    
    struct smbc_dirent *ret;
    SMBC_READDIR_INFO readdir_info;
    SMBC_ACK_INFO ack_info;
    SMBC_DIR_HANDLE *dir;

#ifdef SMBC_USE_SHARED_MEMORY
    dir = (SMBC_DIR_HANDLE *)fd;
    if (dir->sig != 'SMBD') {
	printf("smbc_readdir: Invalid dir handle %d\n", fd);
	return -1;
    }
    fd = dir->fd;
#endif

    smbc_enter_mutex();
    readdir_info.fd       = fd;
#ifdef SMBC_USE_SHARED_MEMORY
    readdir_info.h        = smbc_shared_h;
#endif
    readdir_info.hdr.type = SMBC_MSG_READDIR;
    readdir_info.hdr.size = sizeof(SMBC_READDIR_INFO);
    ret = smbc_send_recv_msg(&readdir_info, &ack_info) ? ack_info.value.pvValue : NULL;
#ifdef SMBC_USE_SHARED_MEMORY
    if (ret) {
	ret = g_smbc_shmaps[readdir_info.h].base_address;
	if (ret->dirlen < sizeof(dir->dirent)) {
	    memcpy(dir->dirent, ret, ret->dirlen);
	    ret = dir->dirent;
	}
	else {
	    printf("smbc_readdir: Dirent overflow %d\n", ret->dirlen);
	    ret = NULL;
	}
    }
#endif
    smbc_leave_mutex();
    return ret;
}

long long smbc_telldir(int dh)
{
    long long ret;
    SMBC_TELLDIR_INFO telldir_info;
    SMBC_ACK_INFO ack_info;

    smbc_enter_mutex();
    telldir_info.dh       = dh;
    telldir_info.hdr.type = SMBC_MSG_TELLDIR;
    telldir_info.hdr.size = sizeof(SMBC_TELLDIR_INFO);
    ret = smbc_send_recv_msg(&telldir_info, &ack_info) ? ack_info.value.llValue : -1;
    smbc_leave_mutex();
    return ret;
}

int smbc_lseekdir(int fd, long long offset)
{
    int ret;
    SMBC_LSEEKDIR_INFO lseekdir_info;
    SMBC_ACK_INFO ack_info;

    smbc_enter_mutex();
    lseekdir_info.fd       = fd;
    lseekdir_info.offset   = offset;
    lseekdir_info.hdr.type = SMBC_MSG_LSEEKDIR;
    lseekdir_info.hdr.size = sizeof(SMBC_LSEEKDIR_INFO);
    ret = smbc_send_recv_msg(&lseekdir_info, &ack_info) ? ack_info.value.iValue : -1;
    smbc_leave_mutex();
    return ret;
}

int smbc_mkdir(const char *fname, mode_t mode)
{
    int ret;
    int i = 0;
    SMBC_MKDIR_INFO mkdir_info;
    SMBC_ACK_INFO ack_info;

    smbc_enter_mutex();
    mkdir_info.fname    = fname;
    mkdir_info.mode     = mode;
    mkdir_info.hdr.type = SMBC_MSG_MKDIR;
    mkdir_info.hdr.size = sizeof(SMBC_MKDIR_INFO);
#ifdef SMBC_USE_SHARED_MEMORY
    mkdir_info.h        = smbc_shared_h;
    mkdir_info.fname    = smbc_wire_string(mkdir_info.h, &i, mkdir_info.fname);
#endif
    ret = smbc_send_recv_msg(&mkdir_info, &ack_info) ? ack_info.value.iValue : -1;
    smbc_leave_mutex();
    return ret;
}

int smbc_rmdir(const char *durl)
{
    int ret;
    int i = 0;
    SMBC_RMDIR_INFO rmdir_info;
    SMBC_ACK_INFO ack_info;

    smbc_enter_mutex();
    rmdir_info.durl     = durl;
    rmdir_info.hdr.type = SMBC_MSG_RMDIR;
    rmdir_info.hdr.size = sizeof(SMBC_RMDIR_INFO);
#ifdef SMBC_USE_SHARED_MEMORY
    rmdir_info.h        = smbc_shared_h;
    rmdir_info.durl     = smbc_wire_string(rmdir_info.h, &i, rmdir_info.durl);
#endif
    ret = smbc_send_recv_msg(&rmdir_info, &ack_info) ? ack_info.value.iValue : -1;
    smbc_leave_mutex();
    return ret;
}

int smbc_stat(const char *fname, struct stat *st)
{
    int ret;
    int i = 0;
    SMBC_STAT_INFO stat_info;
    SMBC_ACK_INFO ack_info;

    SMBC_DEBUG(("smbc_stat\n"));

    smbc_enter_mutex();
    stat_info.fname     = fname;
    stat_info.st        = st;
    stat_info.hdr.type  = SMBC_MSG_STAT;
    stat_info.hdr.size  = sizeof(SMBC_STAT_INFO);
#ifdef SMBC_USE_SHARED_MEMORY
    stat_info.h         = smbc_shared_h;
    stat_info.fname     = smbc_wire_string(stat_info.h, &i, stat_info.fname);
    stat_info.st        = i;
#endif
    ret = smbc_send_recv_msg(&stat_info, &ack_info) ? ack_info.value.iValue : -1;
#ifdef SMBC_USE_SHARED_MEMORY
    if (ret >= 0)
	smbc_unwire_data(stat_info.h, (int)stat_info.st, st, sizeof(struct stat));
#endif
    smbc_leave_mutex();
    return ret;
}

int smbc_fstat(int fd, struct stat *st)
{
    int ret;
    SMBC_FSTAT_INFO fstat_info;
    SMBC_ACK_INFO ack_info;

    SMBC_DEBUG(("smbc_fstat\n"));

    smbc_enter_mutex();
    fstat_info.fd       = fd;
    fstat_info.st       = st;
    fstat_info.hdr.type = SMBC_MSG_FSTAT;
    fstat_info.hdr.size = sizeof(SMBC_FSTAT_INFO);
#ifdef SMBC_USE_SHARED_MEMORY
    fstat_info.st       = 0;
#endif
    ret = smbc_send_recv_msg(&fstat_info, &ack_info) ? ack_info.value.iValue : -1;
#ifdef SMBC_USE_SHARED_MEMORY
    if (ret >= 0)
	smbc_unwire_data(fstat_info.h, (int)fstat_info.st, st, sizeof(struct stat));
#endif
    smbc_leave_mutex();
    return ret;
}

#if 0
int smbc_chown(const char *url, uid_t owner, gid_t group)
{
    int ret;
    int i = 0;
    SMBC_CHOWN_INFO chown_info;
    SMBC_ACK_INFO ack_info;

    smbc_enter_mutex();
    chown_info.url      = url;
    chown_info.owner    = owner;
    chown_info.group    = group;
    chown_info.hdr.type = SMBC_MSG_CHOWN;
    chown_info.hdr.size = sizeof(SMBC_CHOWN_INFO);
#ifdef SMBC_USE_SHARED_MEMORY
    chown_info.h        = smbc_shared_h;
    chown_info.url      = smbc_wire_string(chown_info.h, &i, chown_info.url);
#endif
    ret = smbc_send_recv_msg(&chown_info, &ack_info) ? ack_info.value.iValue : -1;
    smbc_leave_mutex();
    return ret;
}

int smbc_chmod(const char *url, mode_t mode)
{
    int ret;
    int i = 0;
    SMBC_CHMOD_INFO chmod_info;
    SMBC_ACK_INFO ack_info;

    smbc_enter_mutex();
    chmod_info.url      = url;
    chmod_info.mode     = mode;
    chmod_info.hdr.type = SMBC_MSG_CHMOD;
    chmod_info.hdr.size = sizeof(SMBC_CHMOD_INFO);
#ifdef SMBC_USE_SHARED_MEMORY
    chmod_info.h        = smbc_shared_h;
    chmod_info.url      = smbc_wire_string(chmod_info.h, &i, chomod_info.url);
#endif
    ret = smbc_send_recv_msg(&chmod_info, &ack_info) ? ack_info.value.iValue : -1;
    smbc_leave_mutex();
    return ret;
}
#endif

void smbc_remove_unused_servers(unsigned long msec)
{
    SMBC_REMOVEUNUSEDSERVERS_INFO rus_info;
    SMBC_ACK_INFO ack_info;

    smbc_enter_mutex();
    rus_info.msec     = msec;
    rus_info.hdr.type = SMBC_MSG_REMOVEUNUSEDSERVERS;
    rus_info.hdr.size = sizeof(SMBC_REMOVEUNUSEDSERVERS_INFO);
    smbc_send_recv_msg(&rus_info, &ack_info);
    smbc_leave_mutex();
}

int nmbc_name_query(const char *name, struct in_addr bcast_addr, struct in_addr *ip_addr)
{
    int ret;
    int i = 0;
    NMBC_NAMEQUERY_INFO nq_info;
    SMBC_ACK_INFO ack_info;

    SMBC_DEBUG(("smbc_name_query\n"));

    smbc_enter_mutex();
    nq_info.name     = name;
    nq_info.ip_addr  = ip_addr;
    memcpy(&nq_info.bcast_addr, &bcast_addr, sizeof(struct in_addr));
    nq_info.hdr.type = NMBC_MSG_NAMEQUERY;
    nq_info.hdr.size = sizeof(NMBC_NAMEQUERY_INFO);
#ifdef SMBC_USE_SHARED_MEMORY
    nq_info.h        = smbc_shared_h;
    nq_info.name     = smbc_wire_string(nq_info.h, &i, nq_info.name);
    nq_info.ip_addr  = i;
#endif
    ret = smbc_send_recv_msg(&nq_info, &ack_info) ? ack_info.value.iValue : -1;
#ifdef SMBC_USE_SHARED_MEMORY
    if (ret >= 0)
	smbc_unwire_data(nq_info.h, (int)nq_info.ip_addr, ip_addr, sizeof(struct in_addr));
#endif
    smbc_leave_mutex();
    return ret;
}

#ifdef SMBC_USE_SHARED_MEMORY
int smbc_alloc_shared(int size, void **ppv)
{
    int h;
    SMBC_ALLOCSHARED_INFO allocsh_info;
    SMBC_FREESHARED_INFO freesh_info;
    SMBC_ACK_INFO ack_info;
    int i;

    smbc_enter_mutex();
    allocsh_info.size = size;
    allocsh_info.hdr.type = SMBC_MSG_ALLOCSHARED;
    allocsh_info.hdr.size = sizeof(SMBC_ALLOCSHARED_INFO);
    h = smbc_send_recv_msg(&allocsh_info, &ack_info) ? ack_info.value.iValue : -1;
    if (h >= 0) {

	g_smbc_shmaps[h].id = shmget('SMB0'+h, size, 0666 | IPC_CREAT);
	if (g_smbc_shmaps[h].id >= 0) {
	    g_smbc_shmaps[h].base_address = shmat(g_smbc_shmaps[h].id, NULL, 0);
	    if (g_smbc_shmaps[h].base_address) {
		g_smbc_shmaps[h].size = size;
		*ppv = g_smbc_shmaps[h].base_address;
		smbc_leave_mutex();
		return h;
	    }
	    shmctl(g_smbc_shmaps[h].id, IPC_RMID, 0);
	    g_smbc_shmaps[h].id = -1;
	}

	freesh_info.h = h;
	freesh_info.hdr.type = SMBC_MSG_FREESHARED;
	freesh_info.hdr.size = sizeof(SMBC_FREESHARED_INFO);
	smbc_send_recv_msg(&freesh_info, &ack_info);
	h = -1;
    }
    smbc_leave_mutex();
    *ppv = NULL;
    return h;
}

int smbc_free_shared(int h)
{
    SMBC_FREESHARED_INFO freesh_info;
    SMBC_ACK_INFO ack_info;

    smbc_enter_mutex();
    if (h >= 0 && h < arrayof(g_smbc_shmaps) && g_smbc_shmaps[h].base_address) {

	shmdt(g_smbc_shmaps[h].base_address);
	g_smbc_shmaps[h].base_address = NULL;
	
	if (g_smbc_shmaps[h].id >= 0) {
	    shmctl(g_smbc_shmaps[h].id, IPC_RMID, 0);
	    g_smbc_shmaps[h].id = -1;
	}


	freesh_info.h = h;
	freesh_info.hdr.type = SMBC_MSG_FREESHARED;
	freesh_info.hdr.size = sizeof(SMBC_FREESHARED_INFO);
	h = smbc_send_recv_msg(&freesh_info, &ack_info) ? ack_info.value.iValue : -1;
    }
    else {
	printf("smbc_free_shared: Invalid shared handle\n");
	h = -1;
    }

    smbc_leave_mutex();
    return h;
}

int smbc_read_shared(int fd, void *buf, size_t count, int h)
{
    int ret = -1;
    SMBC_READ_INFO read_info;
    SMBC_ACK_INFO ack_info;

    SMBC_DEBUG(("smbc_read_shared\n"));

    smbc_enter_mutex();

    if (h >= 0 && h < arrayof(g_smbc_shmaps) && g_smbc_shmaps[h].base_address) {

	if (buf >= g_smbc_shmaps[h].base_address && (char *)buf+count <= (char *)g_smbc_shmaps[h].base_address + g_smbc_shmaps[h].size) {

	    read_info.fd       = fd;
	    read_info.count    = count;
	    read_info.buf      = (char *)buf - (char *)g_smbc_shmaps[h].base_address;
	    read_info.h        = h;
	    read_info.hdr.type = SMBC_MSG_READ;
	    read_info.hdr.size = sizeof(SMBC_READ_INFO);
	    ret = smbc_send_recv_msg(&read_info, &ack_info) ? ack_info.value.iValue : -1;
	}
	else {
	    printf("smbc_read_shared: Address out of range of shared buffer\n");
	}
    }
    else {
	printf("smbc_read_shared: Invalid shared handle\n");
    }

    smbc_leave_mutex();
    return ret;
}

int smbc_write_shared(int fd, void *buf, size_t count, int h)
{
    int ret = -1;
    SMBC_WRITE_INFO write_info;
    SMBC_ACK_INFO ack_info;

    smbc_enter_mutex();

    if (h >= 0 && h < arrayof(g_smbc_shmaps) && g_smbc_shmaps[h].base_address) {

	if (buf >= g_smbc_shmaps[h].base_address && (char *)buf+count <= (char *)g_smbc_shmaps[h].base_address + g_smbc_shmaps[h].size) {

	    write_info.fd       = fd;
	    write_info.count    = count;
	    write_info.buf      = (char *)buf - (char *)g_smbc_shmaps[h].base_address;
	    write_info.h        = h;
	    write_info.hdr.type = SMBC_MSG_WRITE;
	    write_info.hdr.size = sizeof(SMBC_WRITE_INFO);
	    ret = smbc_send_recv_msg(&write_info, &ack_info) ? ack_info.value.iValue : -1;
	}
	else {
	    printf("smbc_write_shared: Address out of range of shared buffer\n");
	}
    }
    else {
	printf("smbc_write_shared: Invalid shared handle\n");
    }

    smbc_leave_mutex();
    return ret;
}

int smbc_wire_string(int h, int *pi, char *str)
{
    char *sbuf;
    int i, j;

    if (!pi)
	return -1;
    i = *pi;

    //printf("smbc_wire_string, str = %s\n", str ? str : "(null)");

    if (!str) {
	printf("smbc_wire_string: null pointer!\n");
	return -1;
    }

    j = strlen(str) + 1;
    if (i + j > g_smbc_shmaps[h].size) {
	printf("smbc_wire_string: Buffer overrun %d + %d > %d\n", i, j, g_smbc_shmaps[h].size);
	return -1;
    }
    sbuf = (char *)g_smbc_shmaps[h].base_address + i;
    strcpy(sbuf, str);
    (*pi) += j;
    return i;
}

int smbc_unwire_data(int h, int i, void *buf, int len)
{
    char *sbuf = (char *)g_smbc_shmaps[h].base_address + i;
    memcpy(buf, sbuf, len);
    return len;
}
#endif

void smbc_send_recv_sighandler(interruptible_t *pintr)
{
    printf("[%d] smbc_send_recv_sighandler\n", getpid());
    if (pintr->signalled)
	interrupt_pid(smbc_client_pid);
}

int smbc_send_recv_msg(SMBC_MSG *msg, SMBC_ACK_INFO *ack_msg)
{
    int ret;
    interruptible_t intr;

    intr.handler = NULL;
    if (msg->type != SMBC_MSG_CLOSE && msg->type != SMBC_MSG_CLOSEDIR) {

	intr.handler = smbc_send_recv_sighandler;
	intr.refdata = &intr;
	if (!enter_interruptible(&intr))
	    return 0;
    }

    ret = smbc_send_msg(msg);
    if (ret) {
	ret = smbc_recv_msg(ack_msg, sizeof(SMBC_ACK_INFO));
	if (ack_msg->_errno)
	    errno = ack_msg->_errno;
    }

    if (intr.handler) {
	leave_interruptible(&intr);
	if (intr.signalled)
	    return 0;
    }

    return ret;
}

int smbc_send_msg(SMBC_MSG *msg)
{
    int i, j;

    //printf("smbc_send_msg: Sending msg %d\n", msg->type);
    for (i = 0; i < msg->size; i += j) {

	j = send(smbc_fd, ((char *)msg)+i, msg->size - i, 0);
	if (j < 0) {
	    printf("smbc_send_msg: Failed to send %d, error %d\n", j, errno);
	    return 0;
	}
    }
    return 1;
}

int smbc_recv_msg(SMBC_MSG *msg, int size)
{
    int i, j;

    //printf("smbc_recv_msg: Receive msg into %d\n", size);
    for (i = 0; i < sizeof(SMBC_MSG); i += j) {

	j = recv(smbc_fd, ((char *)msg) + i, sizeof(SMBC_MSG) - i, 0);
	if (j <= 0) {
	    printf("smbc_recv_msg: Failed to receive %d, error %d\n", j, errno);
	    return 0;
	}
    }
    //printf("smbc_recv_msg: Receive msg %d size %d\n", msg->type, msg->size);

    if (msg->size > size) {
	printf("smbc_recv_msg: Message too large %d\n", msg->size);
	return 0;
    }

    for (i; i < msg->size; i+= j) {

	j = recv(smbc_fd, ((char *)msg) + i, msg->size - i, 0);
	if (j <= 0) {
	    printf("smbc_recv_msg: Failed to receive %d, error %d\n", j, errno);
	    return 0;
	}
    }
    //printf("smbc_recv_msg: Read %d bytes\n", i);
    return i;
}

char *smbc_strcpyn(char *dst, char *src, int len)
{
    char *ptr = dst;

    if (!len)
	return 0;

    while (*src) {

	if (len <= 1) {

	    *ptr = 0;
	    return dst;
	}

	*ptr = *src;
	ptr++;
	src++;
	len--;
    }

    *ptr = 0;
    return dst;
}
