/*
 *  drivers/mtd/nand.c
 *
 *  Overview:
 *   This is the generic MTD driver for NAND flash devices. It should be
 *   capable of working with almost all NAND chips currently available.
 *   
 *	Additional technical information is available on
 *	http://www.linux-mtd.infradead.org/tech/nand.html
 *	
 *  Copyright (C) 2000 Steven J. Hill (sjhill@cotw.com)
 * 		  2002 Thomas Gleixner (tglx@linutronix.de)
 *
 *  10-29-2001  Thomas Gleixner (tglx@linutronix.de)
 * 		- Changed nand_chip structure for controlline function to
 *		support different hardware structures (Access to
 *		controllines ALE,CLE,NCE via hardware specific function. 
 *		- exit out of "failed erase block" changed, to avoid
 *		driver hangup
 *		- init_waitqueue_head added in function nand_scan !!
 *
 *  01-30-2002  Thomas Gleixner (tglx@linutronix.de)
 *		change in nand_writev to block invalid vecs entries
 *
 *  02-11-2002  Thomas Gleixner (tglx@linutronix.de)
 *		- major rewrite to avoid duplicated code
 *		  common nand_write_page function  
 *		  common get_chip function 
 *		- added oob_config structure for out of band layouts
 *		- write_oob changed for partial programming
 *		- read cache for faster access for subsequent reads
 *		from the same page.
 *		- support for different read/write address
 *		- support for device ready/busy line
 *		- read oob for more than one page enabled
 *
 *  02-27-2002	Thomas Gleixner (tglx@linutronix.de)
 *		- command-delay can be programmed
 *		- fixed exit from erase with callback-function enabled
 *
 *  03-21-2002  Thomas Gleixner (tglx@linutronix.de)
 *		- DEBUG improvements provided by Elizabeth Clarke 
 *		(eclarke@aminocom.com)
 *		- added zero check for this->chip_delay
 *
 *  04-03-2002  Thomas Gleixner (tglx@linutronix.de)
 *		- added added hw-driver supplied command and wait functions
 *		- changed blocking for erase (erase suspend enabled)
 *		- check pointers before accessing flash provided by
 *		John Hall (john.hall@optionexist.co.uk)
 *
 *  04-09-2002  Thomas Gleixner (tglx@linutronix.de)
 *		- nand_wait repaired
 *
 *  04-28-2002  Thomas Gleixner (tglx@linutronix.de)	
 *		- OOB config defines moved to nand.h 
 *
 *  08-01-2002  Thomas Gleixner (tglx@linutronix.de)	
 *		- changed my mailaddress, added pointer to tech/nand.html
 *
 *  08-07-2002 	Thomas Gleixner (tglx@linutronix.de)
 *		forced bad block location to byte 5 of OOB, even if
 *		CONFIG_MTD_NAND_ECC_JFFS2 is not set, to prevent
 *		erase /dev/mtdX from erasing bad blocks and destroying
 *		bad block info
 *
 *  08-10-2002 	Thomas Gleixner (tglx@linutronix.de)
 *		Fixed writing tail of data. Thanks to Alice Hennessy
 *		<ahennessy@mvista.com>.
 *
 *  08-10-2002 	Thomas Gleixner (tglx@linutronix.de)
 *		nand_read_ecc and nand_write_page restructured to support
 *		hardware ECC. Thanks to Steven Hein (ssh@sgi.com)
 *		for basic implementation and suggestions.
 *		3 new pointers in nand_chip structure:
 *		calculate_ecc, correct_data, enabled_hwecc 					 
 *		forcing all hw-drivers to support page cache
 *		eccvalid_pos is now mandatory
 *
 *  08-17-2002	tglx: fixed signed/unsigned missmatch in write.c
 *		Thanks to Ken Offer <koffer@arlut.utexas.edu> 	
 *
 *  08-29-2002  tglx: use buffered read/write only for non pagealigned 
 *		access, speed up the aligned path by using the fs-buffer
 *		reset chip removed from nand_select(), implicit done
 *		only, when erase is interrupted
 *		waitfuntion use yield, instead of schedule_timeout
 *		support for 6byte/512byte hardware ECC
 *		read_ecc, write_ecc extended for different oob-layout
 *		selections: Implemented NAND_NONE_OOB, NAND_JFFS2_OOB,
 *		NAND_YAFFS_OOB. fs-driver gives one of these constants
 *		to select the oob-layout fitting the filesystem.
 *		oobdata can be read together with the raw data, when
 *		the fs-driver supplies a big enough buffer.
 *		size = 12 * number of pages to read (256B pagesize)
 *		       24 * number of pages to read (512B pagesize)
 *		the buffer contains 8/16 byte oobdata and 4/8 byte
 *		returncode from calculate_ecc
 *		oobdata can be given from filesystem to program them
 *		in one go together with the raw data. ECC codes are
 *		filled in at the place selected by oobsel.
 *
 *  09-04-2002  tglx: fixed write_verify (John Hall (john.hall@optionexist.co.uk))
 *
 * $Id: dmc_nand.c,v 1.1.1.1 2006/08/16 19:23:02 jdaub Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ChangLog:
 *     09-Nov-2002 Lineo Japan, Inc.  add NAND page read cache
 *     01-Nov-2002 Lineo Japan, Inc.  retry NAND_CMD_SEQIN at calculate_ecc() error
 *     29-Oct-2002 Lineo Japan, Inc.  modify calculate_ecc return type
 *				      emit NAND_CMD_RESET command at I/O error
 *     27-Sep-2002 Lineo Japan, Inc.  modify local variable declarations
 *     26-Sep-2002 Lineo Japan, Inc.  take hamming distance into consideration
 *     20-Sep-2002 Lineo Japan, Inc.  modify error message
 *				      FL_ERASING waits until FL_READY
 *     19-Sep-2002 Lineo Japan, Inc.  modify nand_command() return type
 *				      add erase-by-force mode
 *     18-Sep-2002 Lineo Japan, Inc.  add dev_ready() call after read
 *     17-Sep-2002 Lineo Japan, Inc.  add code for post-badblock
 */

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/pci.h>
#include <linux/vmalloc.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
//#include <linux/mtd/nand_ids.h>
#include <linux/mtd/nand_ecc.h>
#include <linux/mtd/partitions.h>
#include <linux/interrupt.h>
#include <linux/proc_fs.h>
#include <asm/io.h>
#include <asm/tx7901/tms7901.h>
#include <asm/tx7901/tx7901.h>

#undef DEBUG_PROC
#define NAND_BUSY_TIMEOUT	1000000

//#define F_CACHE_DBG

//#include "measure.h"

#define DEV_NAME	"dmc_nand"
//#define NAND_USE_TIMER

/*
 * Macros for low-level register control
 */
#define nand_select()	
#define nand_deselect() 

typedef volatile u32 vu32;

extern int nand_size, nand_start;

static struct nand_oobinfo nand_oob_16 = {
	.useecc = MTD_NANDECC_AUTOPLACE,
	.eccbytes = 6,
	.eccpos = {0, 1, 2, 3, 6, 7},
	.oobfree = { {8, 8} }
};

#if 0
/*
 * out of band configuration for different filesystems
 */
static int oobconfigs[][6] = {
	{ 0,0,0,0,0,0},

	{ NAND_JFFS2_OOB_ECCPOS0, NAND_JFFS2_OOB_ECCPOS1, NAND_JFFS2_OOB_ECCPOS2,
	NAND_JFFS2_OOB_ECCPOS3, NAND_JFFS2_OOB_ECCPOS4, NAND_JFFS2_OOB_ECCPOS5 },

	{ NAND_YAFFS_OOB_ECCPOS0, NAND_YAFFS_OOB_ECCPOS1, NAND_YAFFS_OOB_ECCPOS2,
	NAND_YAFFS_OOB_ECCPOS3, NAND_YAFFS_OOB_ECCPOS4, NAND_YAFFS_OOB_ECCPOS5 }	
};
#endif

static struct mtd_partition nand_part_info[] = {
  {
    name: "DMC NAND part1",
    offset: ((2*1024*1024)+(1024*1024/2)),
    //size: (2*1024*1024)
    size: (16*1024*1024)-((2*1024*1024)+(1024*1024/2))
  }
};
#define NAND_NPARTS	1

/*
 * NAND low-level MTD interface functions
 */
static int nand_read (struct mtd_info *mtd, loff_t from, size_t len, size_t * retlen, u_char * buf);
static int nand_read_ecc (struct mtd_info *mtd, loff_t from, size_t len,
			  size_t * retlen, u_char * buf, u_char * eccbuf, struct nand_oobinfo *oobsel);
static int nand_read_oob (struct mtd_info *mtd, loff_t from, size_t len, size_t * retlen, u_char * buf);
static int nand_write (struct mtd_info *mtd, loff_t to, size_t len, size_t * retlen, const u_char * buf);
static int nand_write_ecc (struct mtd_info *mtd, loff_t to, size_t len,
			   size_t * retlen, const u_char * buf, u_char * eccbuf, struct nand_oobinfo *oobsel);
static int nand_write_oob (struct mtd_info *mtd, loff_t to, size_t len, size_t * retlen, const u_char *buf);
static int nand_writev (struct mtd_info *mtd, const struct kvec *vecs,
			unsigned long count, loff_t to, size_t * retlen);
static int nand_writev_ecc (struct mtd_info *mtd, const struct kvec *vecs,
			unsigned long count, loff_t to, size_t * retlen, u_char *eccbuf, struct nand_oobinfo *oobsel);
static int nand_erase (struct mtd_info *mtd, struct erase_info *instr);
static void nand_sync (struct mtd_info *mtd);
static int nand_write_page (struct mtd_info *mtd, struct nand_chip *this, int page, int col, 
			int last, u_char *oob_buf, struct nand_oobinfo *oobsel);

//---------------------------------------------------------------------------------------------
// DMC SPI NAND interface
//---------------------------------------------------------------------------------------------

#define DMC_FPGA_VENDOR_ID				0x10ee
#define DMC_FPGA_DEVICE_ID				0x300
#define FPGA_CLOCK_OFFSET 				0x18

unsigned long FPGA_io_base_address;
vu32 *	      FPGA_CLOCK;

unsigned long g_NumErases = 0;
unsigned long g_NumWrites = 0;
unsigned long g_NumReads  = 0;
unsigned long g_EraseTime = 0;
unsigned long g_WriteTime = 0;
unsigned long g_ReadTime  = 0;
int g_UseCache = 0;

#ifdef NAND_USE_TIMER
#define NAND_START_TIMER() \
{ unsigned long _start = *FPGA_CLOCK, _tdiff;

#define NAND_STOP_TIMER(_v) \
  _tdiff = *FPGA_CLOCK - _start; if (_tdiff < 0x7fffffff) _v += _tdiff/33333; else _v -= _tdiff/33333; }
#else
#define NAND_START_TIMER()
#define NAND_STOP_TIMER(_v)
#endif

#define DMC_CPLD_NOP	0x00
#define DMC_CPLD_S_CLE	0x10
#define DMC_CPLD_S_ALE	0x20		//Set address
#define DMC_CPLD_S_RE	0x30		//Read
#define DMC_CPLD_S_WP	0x40		//Assert Write Protect
#define DMC_CPLD_R_WP	0x50		//remove write protect
#define DMC_CPLD_S_PGM	0x60
#define DMC_CPLD_R_PGM	0x70
#define DMC_CPLD_S_FPGA	0x80
#define DMC_CPLD_S_NAND	0x90
#define DMC_CPLD_S_N2_SEL	0xA0
#define DMC_CPLD_R_N2_SEL	0xB0

#define DMC_NAND_SIZE			(8 << 20)
#define DMC_NAND_BLOCK_SIZE		(8 << 10)
#define DMC_NAND_BLOCK_MASK		(DMC_NAND_BLOCK_SIZE - 1)
#define DMC_NAND_LINE_SIZE		(512)
#define DMC_NAND_LINE_SIZE_LONG	        (512+16)
#define DMC_NAND_LINE_MASK		(DMC_NAND_LINE_SIZE - 1)
#define DMC_NAND_NBAD			(10)

#define DMC_SPI_CPLD_NULL					0
#define DMC_SPI_CPLD_BUS					1
#define DMC_SPI_CPLD_CMD					2

/////////////////////////////////////////////////////////////////////////////////////////////////////
//==== DMC NAND CACHE
typedef struct _NAND_CACHE_ENTRY {
	struct _NAND_CACHE_ENTRY *next_lru;
	struct _NAND_CACHE_ENTRY *prev_lru;
	struct _NAND_CACHE_ENTRY *next;
	struct _NAND_CACHE_ENTRY *prev;
	unsigned long		  page;
	unsigned char		  data[DMC_NAND_LINE_SIZE];
	unsigned char		  oob_data[16];
} NAND_CACHE_ENTRY;

#define NAND_CACHE_BLOCK_SIZE (64*1024)
#define NAND_CACHE_NODE_COUNT ((16384*sizeof(NAND_CACHE_ENTRY))/NAND_CACHE_BLOCK_SIZE + 1)
#define NAND_ENTRIES_IN_BLOCK (NAND_CACHE_BLOCK_SIZE/sizeof(NAND_CACHE_ENTRY))
#define NAND_HASH_SIZE	257
NAND_CACHE_ENTRY *g_NAND_Cache[NAND_HASH_SIZE];
NAND_CACHE_ENTRY *g_NAND_FreeList = 0;
NAND_CACHE_ENTRY *g_NAND_LRUList = 0;
void* g_NAND_Alloc[NAND_CACHE_NODE_COUNT] ={0};
unsigned long	  g_NAND_CacheHits = 0;
unsigned long	  g_NAND_CacheMisses = 0;
unsigned long     g_NAND_CachePages = 0;
unsigned long	  g_NAND_CacheBlocks = 0;

static int nandCacheInit(void)
{
	unsigned long i, n;

#if 0
	NAND_CACHE_ENTRY *p;
	for (n = 0; n < NAND_CACHE_NODE_COUNT; n++) {
		
		g_NAND_Alloc[n] = vmalloc(NAND_CACHE_BLOCK_SIZE);
		if (!g_NAND_Alloc[n])
			break;
		p = (NAND_CACHE_ENTRY *)g_NAND_Alloc[n];
		for(i = 0; i < NAND_ENTRIES_IN_BLOCK; i++){
			p->next = g_NAND_FreeList;
			g_NAND_FreeList = p;
			p++;
		}
		g_NAND_CachePages +=NAND_ENTRIES_IN_BLOCK;
	}
	i = (n * sizeof(NAND_CACHE_ENTRY)) / 1024;
	printk("nandCacheInit: Allocated %d entries (%d KB)\n", n, i);
#endif
	
	for (i = 0; i < NAND_HASH_SIZE; i++)
		g_NAND_Cache[i] = 0;

	return 1;
}

static int nandCacheAllocBlock(void)
{
	unsigned long i, n;
	NAND_CACHE_ENTRY *p;

	n = g_NAND_CacheBlocks;
	if (n >= NAND_CACHE_NODE_COUNT)
		return 0;

	g_NAND_Alloc[n] = vmalloc(NAND_CACHE_BLOCK_SIZE);
	if (!g_NAND_Alloc[n])
		return 0;
	p = (NAND_CACHE_ENTRY *)g_NAND_Alloc[n];
	for(i = 0; i < NAND_ENTRIES_IN_BLOCK; i++){
		p->next = g_NAND_FreeList;
		g_NAND_FreeList = p;
		p++;
	}
	g_NAND_CachePages +=NAND_ENTRIES_IN_BLOCK;
	g_NAND_CacheBlocks++;
	return 1;
}

static NAND_CACHE_ENTRY *nandCacheLookup(unsigned long page)
{
	unsigned long i;
        NAND_CACHE_ENTRY *p;

	i = page % NAND_HASH_SIZE;
	p = g_NAND_Cache[i];
	while (p) {
		if (p->page == page)
			break;
		p = p->next;
	}

	if (p) {
		g_NAND_CacheHits++;
		if(p->prev_lru){
			(p->prev_lru)->next_lru = p->next_lru;
			if(p->next_lru){
				(p->next_lru)->prev_lru = p->prev_lru;
			}
			p->next_lru = g_NAND_LRUList;
			g_NAND_LRUList->prev_lru = p;
			p->prev_lru = 0;
			g_NAND_LRUList = p;
		}
#ifdef F_CACHE_DBG
		printk("nandCacheLookup, page = %x: succeeded\n", page);
#endif	      
	}
	else {
		g_NAND_CacheMisses++;
#ifdef F_CACHE_DBG	       
		printk("nandCacheLookup, page = %x: failed\n", page);
#endif	       
	}
	return p;
}

static void nandCacheInsert(unsigned long page, unsigned char *data, unsigned char *oob)
{
	NAND_CACHE_ENTRY *p, *iter;
	unsigned long i;

	p = g_NAND_FreeList;
	if (!p && nandCacheAllocBlock())
		p = g_NAND_FreeList;
	if (!p) {
		iter=g_NAND_LRUList;
		while(iter->next_lru){
			iter=iter->next_lru;
		}
		(iter->prev_lru)->next_lru = 0;
		if(!iter->prev){
			for(i = 0; i < NAND_HASH_SIZE; i++){
				if(g_NAND_Cache[i] == iter){
					g_NAND_Cache[i] = iter->next;
					break;
				}
			}
			if(iter->next)
				(iter->next)->prev = 0;
		}else{
			(iter->prev)->next = iter->next;
			if(iter->next)
				(iter->next)->prev = iter->prev;
		}
		p=iter;
		if (!p) {
#ifdef F_CACHE_DBG
			printk("nandCacheInsert: Unable to allocate\n");
#endif
			return;
		}
	}else{
		g_NAND_FreeList = p->next;
	}
#ifdef F_CACHE_DBG
	printk("nandCacheInsert: page = %x, oob = %02x%02x%02x%02x%02x%02x%02x%02x\n",
	       page, oob[0], oob[1], oob[2], oob[3], oob[4], oob[5], oob[6], oob[7]);
#endif
	
	p->page = page;
	memcpy(p->data, data, sizeof(p->data));
	if (oob)
		memcpy(p->oob_data, oob, sizeof(p->oob_data));
	
	i = page % NAND_HASH_SIZE;
	p->next = g_NAND_Cache[i];
	p->next_lru = g_NAND_LRUList;
	if(g_NAND_Cache[i])
		g_NAND_Cache[i]->prev = p;
	if(g_NAND_LRUList)
		g_NAND_LRUList->prev_lru = p;
	g_NAND_Cache[i] = p;
	g_NAND_LRUList = p;
	p->prev = 0;
	p->prev_lru = 0;
}

static void nandCacheRemove(unsigned long page)
{
	unsigned long i;
        NAND_CACHE_ENTRY *p;
#ifdef F_CACHE_DBG
	printk("nandCacheRemove: page = %x\n", page);
#endif
	i = page % NAND_HASH_SIZE;
	p = g_NAND_Cache[i];
	while (p) {
		if (p->page == page)
			break;
		p = p->next;
	}
	if (p) {
		if(p->prev)
			(p->prev)->next = p->next;
		else
			g_NAND_Cache[i] = p->next;
		if(p->next)
			(p->next)->prev = p->prev;
		if(p->next_lru)
			(p->next_lru)->prev_lru = p->prev_lru;
		if(p->prev_lru)
			(p->prev_lru)->next_lru = p->next_lru;
		else
			g_NAND_LRUList = p->next_lru;
		p->next = g_NAND_FreeList;
		g_NAND_FreeList = p;
	}
}

static void nandCacheFree(void)
{
	//	NAND_MEM_NODE *p, *prev;
	int i = 0, x = 0;
	printk(KERN_ERR "Freeing nand Cache\n");
	/*
	p = g_NAND_AllocList;
	while(p){
		prev = p;
		p = p->next;
		kfree(prev->block);
		kfree(prev);
		x++;
		g_NAND_CachePages -=8;
	}
	*/
	
	for(i=0; i< g_NAND_CacheBlocks; i++){
		if(g_NAND_Alloc[i])
			vfree(g_NAND_Alloc[i]);
		g_NAND_CachePages -=NAND_ENTRIES_IN_BLOCK;
	}
	g_NAND_CacheBlocks = 0;
	
	printk(KERN_ERR "Freed total of %d", x);

	return;
}


/////////////////////////////////////////////////////////////////////////////////////////////////////
//==== DMC GPIO values ====                                                                                     3210
static u8 gpoutNull    = 0xf5;   // DMC 10 change ...    10           1111 0101   (RTC = SPI3)     (=SPI1)
static u8 gpoutCpld    = 0xf3;   // NAND I/F     cs01n = 01 => CPLD   1111 0011
static u8 gpoutCpldCmd = 0xf1;   //              cs01n = 00 => CMD    1111 0001
                          // cs0_n = SPI2, cs1_n = SPI1   watch for the bits order !!!!!
                          // RTC uses SPI_SEL3 (Don't need for DMC11 !!!)

#define NAND_CHIP_FROM_PAGE(_page) (((_page) < 16384) ? 1 : 2)
void BridgeSelectNAND(unsigned char Chip_num);

static void spiInitForSrom(void) 
{
    // Deselect all. 
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutNull;
    *(vu32*)(SPI_GPEN + TX7901_SPI_BASE) = 0;          // Output all enable.

    // Change clock rate.
    *(vu32*)(SPI_SECR + TX7901_SPI_BASE) = 0x10;       // TSEI off.
    *(vu32*)(SPI_DDCR + TX7901_SPI_BASE) = 0x38;
    *(vu32*)(SPI_SECR + TX7901_SPI_BASE) = 0x50;       // TSEI on. /8
}

static void spiWriteByte(u8 data)
{
    u8 sesr;

    // Send a byte.
    *(vu32*)(SPI_SEDR + TX7901_SPI_BASE) = data;

    // Wait until completion.
    do {
        sesr = *(vu32*)(SPI_SESR + TX7901_SPI_BASE);
    } while ((sesr & 0x80) == 0);
}

static u8 spiReadByte(void)
{
    // Dummy write to start transfer, and wait for completion.
    spiWriteByte(0);

    return *(vu32*)(SPI_SEDR + TX7901_SPI_BASE);
}

void nandWritable(void) 
{
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpldCmd;
    spiWriteByte(DMC_CPLD_R_WP);
}

void nandProtect(void) 
{
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpldCmd;
    spiWriteByte(DMC_CPLD_S_WP);
}

void nandCommand(int cmd) 
{
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpldCmd;
    spiWriteByte(DMC_CPLD_S_CLE);
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpld;
    spiWriteByte(cmd);
}

int nandIsReady(void)
{
    u32 rdy;

    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpldCmd;
    spiWriteByte(DMC_CPLD_NOP);
    rdy = *(vu32*)(SPI_SEDR + TX7901_SPI_BASE);

    return (rdy & 4) ? 1 : 0;
}

void nandWait(void) 
{
    u32 rdy;

    do {
        *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpldCmd;
        spiWriteByte(DMC_CPLD_NOP);
        rdy = *(vu32*)(SPI_SEDR + TX7901_SPI_BASE);
    } while ((rdy & 4) == 0);
}

u8 nandStatus(void) 
{
    u8 status;

    nandCommand(0x70);

    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpldCmd;
    spiWriteByte(DMC_CPLD_S_RE);
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpld;
    status = spiReadByte();
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutNull;

    return status;
}

u16 nandId(void) 
{
    u8 maker, device;

    nandCommand(0x90);
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpldCmd;
    spiWriteByte(DMC_CPLD_S_ALE);
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpld;
    spiWriteByte(0x00);

    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpldCmd;
    spiWriteByte(DMC_CPLD_S_RE);
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpld;
    maker = spiReadByte();
    device = spiReadByte();
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutNull;

    return (maker << 8) | device;
}

void nandReset(void) 
{
    nandCommand(0xff);
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutNull;
}

void nandErasePage(u32 page) 
{
	//printk("nandErasePage: page %x\n", page);
	u32 i;
	
	if(g_UseCache){
		for (i = 0; i < DMC_NAND_BLOCK_SIZE/DMC_NAND_LINE_SIZE; i++)
			nandCacheRemove(page + i);
	}
	
	NAND_START_TIMER();

    BridgeSelectNAND(NAND_CHIP_FROM_PAGE(page));

    nandWritable();

    nandCommand(0x60);
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpldCmd;
    spiWriteByte(DMC_CPLD_S_ALE);
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpld;
    spiWriteByte(page);
    spiWriteByte(page >> 8);

    nandCommand(0xd0);
    nandWait();
    nandProtect();
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutNull;

    NAND_STOP_TIMER(g_EraseTime);
    g_NumErases++;
}
   
u8 nandCheckPage(u32 page)
{
    u8 data;

    //printk("nandCheckPage: page %x\n", page);

    BridgeSelectNAND(NAND_CHIP_FROM_PAGE(page));

    // Check if each block is good by looking at byte zero of the hidden area.
    // If it is 0xff, we assume it is a good block.
	nandCommand(0x50); // hidden area  (512+ 0 .. 15)
   *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpldCmd;
   spiWriteByte(DMC_CPLD_S_ALE);
   *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpld;
   spiWriteByte(0x00);
   spiWriteByte(page);
   spiWriteByte(page >> 8);
   nandWait();

   *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpldCmd;
   spiWriteByte(DMC_CPLD_S_RE);
   *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpld;
   data = spiReadByte();
   *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutNull;

   return data;
}

void nandReadOOB(u32 page, unsigned char *raddr)
{
   int i;
	
   //printk("nandReadOOB: page = %x\n", page);

   BridgeSelectNAND(NAND_CHIP_FROM_PAGE(page));

   nandCommand(0x50); // hidden area  (512+ 0 .. 15)
   *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpldCmd;
   spiWriteByte(DMC_CPLD_S_ALE);
   *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpld;
   spiWriteByte(0x00);
   spiWriteByte(page);
   spiWriteByte(page >> 8);
   nandWait();

   *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpldCmd;
   spiWriteByte(DMC_CPLD_S_RE);
   *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpld;

   for (i = 0; i < 16; i++)
	   raddr[i] = spiReadByte();
   *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutNull;
}

void nandReadPage(u32 page, unsigned char *raddr, unsigned char *oob)
{
	int i;
	NAND_CACHE_ENTRY *p;

	if(g_UseCache){
		p = nandCacheLookup(page);
		if (p) {
			memcpy(raddr, p->data, sizeof(p->data));
			if (oob)
				memcpy(oob, p->oob_data, sizeof(p->oob_data));
			return;
		}
	}

	//printk("nandReadPage: page = %x\n", page);
	NAND_START_TIMER();

	BridgeSelectNAND(NAND_CHIP_FROM_PAGE(page));

        nandCommand(0x00);
        *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpldCmd;
        spiWriteByte(DMC_CPLD_S_ALE);
        *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpld;
        spiWriteByte(0x00);
        spiWriteByte(page);
        spiWriteByte(page >> 8);
        nandWait();

        *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpldCmd;
        spiWriteByte(DMC_CPLD_S_RE);
        *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpld;

        for (i = 0; i < DMC_NAND_LINE_SIZE; i++) {
		raddr[i] = spiReadByte();
        }
        *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutNull;

	if (oob) 
		nandReadOOB(page, oob);
	if(g_UseCache){
		for (i = 0; i < DMC_NAND_LINE_SIZE; i++)
			if (raddr[i] != 0xFF)
				break;
		if (i < DMC_NAND_LINE_SIZE){
			nandCacheInsert(page, raddr, oob);
		}
	}
	NAND_STOP_TIMER(g_ReadTime);
	g_NumReads++;
}

void nandWritePage(u32 page, unsigned char *raddr, unsigned char *oob)
{
	int i;

	//printk("nandWritePage: page = %x\n", page);
	if (page < 256) {
		printk("nandBlockWrite: About to overwrite FPGA monitor at page %x!\n", page);
		return;
	}		

	if(g_UseCache){
		nandCacheRemove(page);
	}
	NAND_START_TIMER();

	BridgeSelectNAND(NAND_CHIP_FROM_PAGE(page));

        nandWritable();
        nandCommand(0x00);
        spiWriteByte(DMC_CPLD_S_NAND);
        nandCommand(0x80);
        *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpldCmd;
        spiWriteByte(DMC_CPLD_S_ALE);
        *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpld;
        spiWriteByte(0);
        spiWriteByte(page);
        spiWriteByte(page >> 8);

        // To end ALE.
        *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpldCmd;
        spiWriteByte(DMC_CPLD_S_NAND);
        *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpld;

	//debugo(2,"got here\n");
        for (i = 0; i < DMC_NAND_LINE_SIZE; i++) {
		spiWriteByte(raddr[i]);
        }
	
        for (i = 0; i < DMC_NAND_LINE_SIZE_LONG - DMC_NAND_LINE_SIZE; i++) {
		spiWriteByte(oob ? oob[i] : 0xff);
        }
        nandCommand(0x10); // finish the write cycle

        // To end CLE.
        *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpldCmd;
        spiWriteByte(DMC_CPLD_S_NAND);
        *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpld;

        nandWait();
        nandProtect();
        *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutNull;

	NAND_STOP_TIMER(g_ReadTime);
	g_NumWrites++;
}

//==================================================
// FPGA config ...
//==================================================
void BridgeSetNAND(void)
{
    // reset the bridge
    nandCommand(0xff);
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutNull;
	
	// set the bridge to NAND output ...
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpldCmd;
    spiWriteByte(DMC_CPLD_S_NAND);
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpld;
}

void BridgeSetFPGA(void)
{
	// set the bridge output to fpga
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpldCmd;
    spiWriteByte(DMC_CPLD_S_FPGA);
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpld;
}


void BridgeXilinxStartPGM(void)
{
    // reset the bridge
    nandCommand(0xff);
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutNull;

	// start fpga program by pulsing at PGM signal ...
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpldCmd;
    spiWriteByte(DMC_CPLD_S_PGM);
    spiWriteByte(DMC_CPLD_R_PGM);

    spiWriteByte(DMC_CPLD_S_FPGA);
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpld;
}

unsigned char GetBridgeStatus(void)
{
	// start fpga program by pulsing at PGM signal ...
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpldCmd;
    
    // read the status when writing NOP
    spiWriteByte(DMC_CPLD_NOP);
    return *(vu32*)(SPI_SEDR + TX7901_SPI_BASE);
}

void BridgeSelectNAND(unsigned char Chip_num)
{
    static unsigned char m_Chip_num = 0;

    //if (m_Chip_num == Chip_num)
    //return;

    if (m_Chip_num == 0) {
	BridgeSetNAND();
	spiInitForSrom();
    }

    // reset the bridge
    nandCommand(0xff);
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutNull;

    // select the nand chip 1,2
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpldCmd;
    if (Chip_num == 2) {
    	spiWriteByte(DMC_CPLD_S_N2_SEL);
	//printk("dmc_nand: NAND2 selected\n");
    }
    else {
    	spiWriteByte(DMC_CPLD_R_N2_SEL);
	//printk("dmc_nand: NAND1 selected\n");
    }

    spiWriteByte(DMC_CPLD_S_NAND);
    *(vu32*)(SPI_GPOUT + TX7901_SPI_BASE) = gpoutCpld;

    m_Chip_num = Chip_num;
}

//---------------------------------------------------------------------------------------------
// NAND interface
//---------------------------------------------------------------------------------------------

/*
 * Send command to NAND device
 */
static int nand_command (struct mtd_info *mtd, unsigned command, int column, int page_addr)
{
	return 0;
}

static int nand_dev_ready(struct mtd_info *mtd)
{
	return 1;
}

static int nand_block_bad(struct mtd_info *mtd, loff_t ofs, int getchip)
{
	int page;
	struct nand_chip *this = mtd->priv;
	u16 bad;

	page = (int)(ofs >> this->page_shift);
	if (nandCheckPage(page) != 0xff)
		return 1;
	return 0;
}

static int nand_block_checkbad (struct mtd_info *mtd, loff_t ofs, int getchip, int allowbbt)
{
	struct nand_chip *this = mtd->priv;
	
	if (!this->bbt)
		return nand_block_bad(mtd, ofs, getchip);
	
	/* Return info from the table */
	return nand_isbad_bbt (mtd, ofs, allowbbt);
}

static int nand_default_block_markbad(struct mtd_info *mtd, loff_t ofs)
{
	struct nand_chip *this = mtd->priv;
	u_char buf[2] = {0, 0};
	size_t	retlen;
	int block;
	
	/* Get block number */
	block = ((int) ofs) >> this->bbt_erase_shift;
	if (this->bbt)
		this->bbt[block >> 2] |= 0x01 << ((block & 0x03) << 1);

	/* Do we have a flash based bad block table ? */
	if (this->options & NAND_USE_FLASH_BBT)
		return nand_update_bbt (mtd, ofs);
		
	/* We write two bytes, so we dont have to mess with 16 bit access */
	ofs += mtd->oobsize + (this->badblockpos & ~0x01);
	return nand_write_oob (mtd, ofs , 2, &retlen, buf);
}

static int nand_block_isbad (struct mtd_info *mtd, loff_t ofs)
{
	/* Check for invalid offset */
	if (ofs > mtd->size) 
		return -EINVAL;
	
	return nand_block_checkbad (mtd, ofs, 1, 0);
}

static int nand_block_markbad (struct mtd_info *mtd, loff_t ofs)
{
	struct nand_chip *this = mtd->priv;
	int ret;

        if ((ret = nand_block_isbad(mtd, ofs))) {
        	/* If it was bad already, return success and do nothing. */
		if (ret > 0)
			return 0;
        	return ret;
        }

	return nand_default_block_markbad(mtd, ofs);
}

/*
 *	Get chip for selected access
 */
static inline void nand_get_chip (struct nand_chip *this, struct mtd_info *mtd, int new_state, int *erase_state)
{

	DECLARE_WAITQUEUE (wait, current);

	//printk("nand_get_chip, this = %x\n", this);

	/* 
	 * Grab the lock and see if the device is available 
	 * For erasing, we keep the spinlock until the
	 * erase command is written. 
	*/
retry:
	spin_lock_bh (&this->chip_lock);

	if (this->state == FL_READY) {
		this->state = new_state;
		if (new_state != FL_ERASING)
			spin_unlock_bh (&this->chip_lock);
		return;
	}

#if 0
	if (this->state == FL_ERASING) {
		if (new_state != FL_ERASING) {
			this->state = new_state;
			spin_unlock_bh (&this->chip_lock);
			nand_select ();	/* select in any case */
			this->cmdfunc(mtd, NAND_CMD_RESET, -1, -1);
			return;
		}
	}
#endif

	//printk("nand_get_chip: setting interruptible\n");
	set_current_state (TASK_UNINTERRUPTIBLE);
	add_wait_queue (&this->wq, &wait);
	spin_unlock_bh (&this->chip_lock);
	schedule ();
	remove_wait_queue (&this->wq, &wait);
	goto retry;
}

/*
 * Wait for command done. This applies to erase and program only
 * Erase can take up to 400ms and program up to 20ms according to 
 * general NAND and SmartMedia specs
 *
*/
static int nand_wait(struct mtd_info *mtd, struct nand_chip *this, int state)
{

	unsigned long	timeo = jiffies;
	int	status;

	//printk("nand_wait\n");

	if (state == FL_ERASING)
		 timeo += (HZ * 400) / 1000;
	else
		 timeo += (HZ * 20) / 1000;

	spin_lock_bh (&this->chip_lock);

	while (time_before(jiffies, timeo)) {		
		/* Check, if we were interrupted */
		if (this->state != state) {
			spin_unlock_bh (&this->chip_lock);
			return 0;
		}
		if (this->dev_ready) {
			if (this->dev_ready (mtd))
				break;
		}

		spin_unlock_bh (&this->chip_lock);
		yield ();
		spin_lock_bh (&this->chip_lock);
	}
	status = 0;
	spin_unlock_bh (&this->chip_lock);

	return status;
}

/*
 *	Nand_page_program function is used for write and writev !
 */
static int nand_write_page (struct mtd_info *mtd, struct nand_chip *this,
			    const int page, int col, int last, u_char *oob_buf, struct nand_oobinfo *oobsel)
{
	int 	i;
	u_char	ecc_code[6], *oob_data;
	int	eccmode = oobsel->useecc ? this->eccmode : NAND_ECC_NONE;
	int  	*oob_config = oobsel->eccpos;
	int	ecc_ret = 0;
	unsigned char data[DMC_NAND_LINE_SIZE_LONG];

	//START_MEASUREMENT(nand_write_page);
	//COUNTER_INC(nand_write_nr_pages);

	//printk("nand_write_page: page = %x\n", page);

	nandReadPage(page, &data[0], &data[DMC_NAND_LINE_SIZE]);

	/* pad oob area, if we have no oob buffer from fs-driver */
	if (!oob_buf) {
		oob_data = &data[DMC_NAND_LINE_SIZE];
		for (i = 0; i < mtd->oobsize; i++)
			oob_data[i] = 0xff;
	} else 
		oob_data = oob_buf;

	for (i = col; i < last; i++)
		data[i] = this->data_poi[i];

	/* software ecc 3 Bytes ECC / 256 Byte Data ? */
	if (eccmode == NAND_ECC_SOFT) {
		/* Read back previous written data, if col > 0 */
		if ((col < this->eccsize) && (last >= this->eccsize)) {
			this->calculate_ecc (mtd, &data[0], &(ecc_code[0]));
			for (i = 0; i < 3; i++)
				oob_data[oob_config[i]] = ecc_code[i];
		}
		/* Calculate and write the second ECC if we have enough data */
		if ((mtd->oobblock == 512) && (last == 512)) {
			this->calculate_ecc (mtd, &data[256], &(ecc_code[3]));
			for (i = 3; i < 6; i++)
				oob_data[oob_config[i]] = ecc_code[i];
		} 
	} else {
		/* For hardware ECC skip ECC, if we have no full page write */
		if (eccmode != NAND_ECC_NONE && (col || last != mtd->oobblock))
			eccmode = NAND_ECC_NONE;
	}			

  ecc_calc_retry:
	/* Send command to begin auto page programming */
	nandWritePage(page, &data[0], oob_data);

	/* Write out complete page of data, take care of eccmode */
	switch (this->eccmode) {
	/* No ecc and software ecc 3/256, write all */
	case NAND_ECC_NONE:
	case NAND_ECC_SOFT:
		break;

#if 0		
	/* Hardware ecc 3 byte / 256 data, write first half, get ecc, then second, if 512 byte pagesize */	
	case NAND_ECC_HW3_256:		
		this->enable_hwecc (NAND_ECC_WRITE);	/* enable hardware ecc logic for write */
		for (i = 0; i < mtd->eccsize; i++) 
			writeb ( this->data_poi[i] , this->IO_ADDR_W);
		
		ecc_ret = this->calculate_ecc (mtd, NULL, &(ecc_code[0]));
		for (i = 0; i < 3; i++)
			oob_data[oob_config[i]] = ecc_code[i];
			
		if (mtd->oobblock == 512) {
			this->enable_hwecc (NAND_ECC_WRITE);	/* enable hardware ecc logic for write*/
			for (i = mtd->eccsize; i < mtd->oobblock; i++) 
				writeb ( this->data_poi[i] , this->IO_ADDR_W);
			ecc_ret |= this->calculate_ecc (mtd, NULL, &(ecc_code[3]));
			for (i = 3; i < 6; i++)
				oob_data[oob_config[i]] = ecc_code[i];
		}
		break;
				
	/* Hardware ecc 3 byte / 512 byte data, write full page */	
	case NAND_ECC_HW3_512:	
		this->enable_hwecc (NAND_ECC_WRITE);	/* enable hardware ecc logic */
		for (i = 0; i < mtd->oobblock; i++) 
			writeb ( this->data_poi[i] , this->IO_ADDR_W);
		ecc_ret = this->calculate_ecc (mtd, NULL, &(ecc_code[0]));
		for (i = 0; i < 3; i++)
			oob_data[oob_config[i]] = ecc_code[i];
		break;

	/* Hardware ecc 6 byte / 512 byte data, write full page */	
	case NAND_ECC_HW6_512:	
		this->enable_hwecc (NAND_ECC_WRITE);	/* enable hardware ecc logic */
		for (i = 0; i < mtd->oobblock; i++) 
			writeb ( this->data_poi[i] , this->IO_ADDR_W);
		ecc_ret = this->calculate_ecc (mtd, NULL, &(ecc_code[0]));
		for (i = 0; i < 6; i++)
			oob_data[oob_config[i]] = ecc_code[i];
		break;
#endif
		
	default:
		printk (KERN_WARNING "Invalid NAND_ECC_MODE %d\n", this->eccmode);
		BUG();	
	}

	if (ecc_ret) {
		DEBUG(MTD_DEBUG_LEVEL0, "%s: Failed write ECC, page 0x%08x, ", __func__, page);
		ecc_ret = 0;
		goto ecc_calc_retry;
	}
	
	//ACCUMULATE_ELAPSED_TIME(nand_write_page);
	//printk("nand_write_page -> 0\n");
	return 0;
}

/*
*	Use NAND read ECC
*/
static int nand_read (struct mtd_info *mtd, loff_t from, size_t len, size_t * retlen, u_char * buf)
{
	return (nand_read_ecc (mtd, from, len, retlen, buf, NULL, NULL));
}			   


/*
 * NAND read with ECC
 */
static int nand_read_ecc (struct mtd_info *mtd, loff_t from, size_t len,
			  size_t * retlen, u_char * buf, u_char * oob_buf, struct nand_oobinfo *oobsel)
{
	int j, col, page, end, ecc;
	int erase_state = 0;
	int read = 0, oob = 0, ecc_status = 0, ecc_failed = 0;
	struct nand_chip *this = mtd->priv;
	u_char *data_poi, *oob_data = oob_buf;
	u_char ecc_calc[6];
	u_char ecc_code[6];
	int	eccmode;
	int *   oob_config;
	unsigned char data[DMC_NAND_LINE_SIZE_LONG];
	
	DEBUG (MTD_DEBUG_LEVEL3, "nand_read_ecc: from = 0x%08x, len = %i\n", (unsigned int) from, (int) len);
	//printk("nand_read_ecc: from = 0x%08x, len = %i, buf = 0x%08x\n", (unsigned int) from, (int) len, buf);

	if (oobsel == NULL)
		oobsel = &mtd->oobinfo;

	if (oobsel->useecc == MTD_NANDECC_AUTOPLACE)
		oobsel = this->autooob;
	
	eccmode = oobsel->useecc ? this->eccmode : NAND_ECC_NONE;
	oob_config = oobsel->eccpos;

	/* Do not allow reads past end of device */
	if ((from + len) > mtd->size) {
		DEBUG (MTD_DEBUG_LEVEL0, "nand_read_ecc: Attempt read beyond end of device\n");
		//printk("nand_read_ecc: Attempt read beyond end\n");
		*retlen = 0;
		return -EINVAL;
	}

	/* Grab the lock and see if the device is available */
	nand_get_chip (this, mtd ,FL_READING, &erase_state);

	/* Select the NAND device */
	nand_select ();

	/* First we calculate the starting page */
	page = from >> this->page_shift;

	/* Get raw starting column */
	col = from & (mtd->oobblock - 1);

	end = mtd->oobblock;
	ecc = mtd->eccsize;

	/* Send the read command */
	//if (this->cmdfunc (mtd, NAND_CMD_READ0, 0x00, page)) {
	//	printk(KERN_WARNING "%s: Failed in NAND_CMD_READ0 command, page 0x%08x\n",
	//	       __func__, page);
	//	ecc_failed = 1;
	//	goto nand_read_ecc_exit;
	//}
	
	/* Loop until all data read */
	while (read < len) {
		/* 
		 * If the read is not page aligned, we have to read into data buffer
		 * due to ecc, else we read into return buffer direct
		 */
		if (!col && (len - read) >= end)  
			data_poi = &buf[read];
		else 
			data_poi = this->data_buf;

		/* get oob area, if we have no oob buffer from fs-driver */
		if (!oob_buf) {
			oob_data = &this->data_buf[end];
			oob = 0;
		} 	

		nandReadPage(page, &data[0], &data[DMC_NAND_LINE_SIZE]);

		j = 0;
		switch (eccmode) {
		case NAND_ECC_NONE:	/* No ECC, Read in a page */		
			while (j < end)
				data_poi[j++] = data[j];
			break;
			
		case NAND_ECC_SOFT:	/* Software ECC 3/256: Read in a page + oob data */
			while (j < end)
				data_poi[j++] = data[j];
			this->calculate_ecc (mtd, &data_poi[0], &ecc_calc[0]);
			if (mtd->oobblock == 512)
				this->calculate_ecc (mtd, &data_poi[256], &ecc_calc[3]);
			break;	
		
#if 0	
		case NAND_ECC_HW3_256: /* Hardware ECC 3 byte /256 byte data: Read in first 256 byte, get ecc, */
			this->enable_hwecc (NAND_ECC_READ);	
			while (j < ecc)
				data_poi[j++] = readb (this->IO_ADDR_R);
			this->calculate_ecc (mtd, &data_poi[0], &ecc_calc[0]);	/* read from hardware */
			
			if (mtd->oobblock == 512) { /* read second, if pagesize = 512 */
				this->enable_hwecc (NAND_ECC_READ);	
				while (j < end)
					data_poi[j++] = readb (this->IO_ADDR_R);
				this->calculate_ecc (mtd, &data_poi[256], &ecc_calc[3]); /* read from hardware */
			}					
			break;						
				
		case NAND_ECC_HW3_512:	
		case NAND_ECC_HW6_512: /* Hardware ECC 3/6 byte / 512 byte data : Read in a page  */
			this->enable_hwecc (NAND_ECC_READ);	
			while (j < end)
				data_poi[j++] = readb (this->IO_ADDR_R);
			this->calculate_ecc (mtd, &data_poi[0], &ecc_calc[0]);	/* read from hardware */
			break;
#endif

		default:
			printk (KERN_WARNING "Invalid NAND_ECC_MODE %d\n", this->eccmode);
			BUG();	
		}

		/* read oobdata */
		for (j = 0; j <  mtd->oobsize; j++) 
			oob_data[oob + j] = data[DMC_NAND_LINE_SIZE + j];
		
		/* Skip ECC, if not active */
		if (eccmode == NAND_ECC_NONE)
			goto readdata;	
		
		/* Pick the ECC bytes out of the oob data */
		for (j = 0; j < 6; j++)
			ecc_code[j] = oob_data[oob + oob_config[j]];

		/* correct data, if neccecary */
		ecc_status = this->correct_data (mtd, &data_poi[0], &ecc_code[0], &ecc_calc[0]);
		/* check, if we have a fs supplied oob-buffer */
		if (oob_buf) { 
			oob += mtd->oobsize;
			*((int *)&oob_data[oob]) = ecc_status;
			oob += sizeof(int);
		}
		if (ecc_status == -1) {	
			printk (KERN_WARNING "%s: Failed ECC read, page 0x%08x\n", __func__, page);
			ecc_failed++;
		}
		
		if (mtd->oobblock == 512 && eccmode != NAND_ECC_HW3_512) {
			ecc_status = this->correct_data (mtd, &data_poi[256], &ecc_code[3], &ecc_calc[3]);
			if (oob_buf) {
				*((int *)&oob_data[oob]) = ecc_status;
				oob += sizeof(int);
			}
			if (ecc_status == -1) {
				printk (KERN_WARNING "%s: Failed ECC read, page 0x%08x\n", __func__, page);
				ecc_failed++;
			}
		}

readdata:
		/* If we have consequent page reads, apply delay or wait for ready/busy pin */
		if (!this->dev_ready) 
			udelay (this->chip_delay);
		else {
			int i;
			for (i = 0; i < NAND_BUSY_TIMEOUT; i++)
				if (this->dev_ready(mtd))
					break;
			if (i == NAND_BUSY_TIMEOUT) {
				ecc_failed++;
				goto nand_read_ecc_exit;
			}
		}

		if (col || (len - read) < end) { 
			for (j = col; j < end && read < len; j++)
				buf[read++] = data_poi[j];
		} else		
			read += mtd->oobblock;
		/* For subsequent reads align to page boundary. */
		col = 0;
		/* Increment page address */
		page++;
	}

  nand_read_ecc_exit:
	if (ecc_failed)
		this->cmdfunc(mtd, NAND_CMD_RESET, -1, -1);

	/* De-select the NAND device */
	nand_deselect ();

	/* Wake up anyone waiting on the device */
	spin_lock_bh (&this->chip_lock);
	this->state = FL_READY;
	wake_up (&this->wq);
	spin_unlock_bh (&this->chip_lock);

	/*
	 * Return success, if no ECC failures, else -EIO
	 * fs driver will take care of that, because
	 * retlen == desired len and result == -EIO
	 */
	*retlen = read;

	//printk("nand_read_ecc -> %d\n", ecc_failed);

	return ecc_failed ? -EIO : 0;
}

/*
 * NAND read out-of-band
 */
static int nand_read_oob (struct mtd_info *mtd, loff_t from, size_t len, size_t * retlen, u_char * buf)
{
	int col, page;
	int erase_state = 0;
	struct nand_chip *this = mtd->priv;
	unsigned char data[DMC_NAND_LINE_SIZE_LONG];

	DEBUG (MTD_DEBUG_LEVEL3, "nand_read_oob: from = 0x%08x, len = %i\n", (unsigned int) from, (int) len);
	//printk("nand_read_oob: from = 0x%08x, len = %i\n", (unsigned int) from, (int) len);

	/* Shift to get page */
	page = ((int) from) >> this->page_shift;

	/* Mask to get column */
	col = from & 0x0f;

	/* Do not allow reads past end of device */
	if ((from + len) > mtd->size) {
		DEBUG (MTD_DEBUG_LEVEL0, "nand_read_oob: Attempt read beyond end of device\n");
		//printk("nand_read_oob: Attempt read beyond end\n");
		*retlen = 0;
		return -EINVAL;
	}

	/* Initialize return length value */
	*retlen = 0;

	/* Grab the lock and see if the device is available */
	nand_get_chip (this, mtd , FL_READING, &erase_state);

	/* Select the NAND device */
	nand_select ();

	while (len) {
		int tmp_len = min(len, mtd->oobsize - col);
		int i;

		nandReadPage(page, &data[0], &data[DMC_NAND_LINE_SIZE]);

		for (i = 0; i < tmp_len; i++) {
			buf[i] = data[DMC_NAND_LINE_SIZE + col + i];
		}

		col = 0;
		page++;
		len -= tmp_len;
		*retlen += tmp_len;
		buf += tmp_len;
	}

	/* De-select the NAND device */
	nand_deselect ();

	/* Wake up anyone waiting on the device */
	spin_lock_bh (&this->chip_lock);
	this->state = FL_READY;
	wake_up (&this->wq);
	spin_unlock_bh (&this->chip_lock);

	/* Return happy */
	//printk("nand_read_oob -> 0\n");
	return 0;
}

/*
*	Use NAND write ECC
*/
static int nand_write (struct mtd_info *mtd, loff_t to, size_t len, size_t * retlen, const u_char * buf)
{
	return (nand_write_ecc (mtd, to, len, retlen, buf, NULL, NULL));
}			   

/*
 * NAND write with ECC
 */
static int nand_write_ecc (struct mtd_info *mtd, loff_t to, size_t len,
			   size_t * retlen, const u_char * buf, u_char * eccbuf, struct nand_oobinfo *oobsel)
{
	int page, col, cnt, ret = 0, oob = 0, written = 0;
	struct nand_chip *this = mtd->priv;

	DEBUG (MTD_DEBUG_LEVEL3, "nand_write_ecc: to = 0x%08x, len = %i\n", (unsigned int) to, (int) len);
	//printk("nand_write_ecc: to = 0x%08x, len = %i\n", (unsigned int) to, (int) len);

	/* Do not allow write past end of device */
	if ((to + len) > mtd->size) {
		DEBUG (MTD_DEBUG_LEVEL0, "nand_write_oob: Attempt to write past end of page\n");
		//printk("nand_write_ecc: Attempt read beyond end\n");
		return -EINVAL;
	}

	/* Shift to get page */
	page = ((int) to) >> this->page_shift;

	/* Get the starting column */
	col = to & (mtd->oobblock - 1);

	/* Grab the lock and see if the device is available */
	nand_get_chip (this, mtd, FL_WRITING, NULL);

	/* Select the NAND device */
	nand_select ();

	/* Check the WP bit */
	//this->cmdfunc (mtd, NAND_CMD_STATUS, -1, -1);
	//if (!(readb (this->IO_ADDR_R) & 0x80)) {
	//	DEBUG (MTD_DEBUG_LEVEL0, "nand_write_ecc: Device is write protected!!!\n");
	//	ret = -EIO;
	//	goto out;
	//}

	/* Loop until all data is written */
	while (written < len) {
		/* 
		 * Check, if we have a full page write, then we can
		 * use the given buffer, else we have to copy
		 */
		if (!col && (len - written) >= mtd->oobblock) {   
			this->data_poi = (u_char*) &buf[written];
			cnt = mtd->oobblock;
		} else {
			cnt = min(len - written, mtd->oobblock);
			memcpy(&this->data_buf[col], &buf[written + col], cnt);
			this->data_poi = this->data_buf;		
		}
		/* We use the same function for write and writev !) */
		if (eccbuf) {
			ret = nand_write_page (mtd, this, page, col, cnt ,&eccbuf[oob], oobsel);
			oob += mtd->oobsize;
		} else 
			ret = nand_write_page (mtd, this, page, col, cnt, NULL, oobsel);	
		
		if (ret)
			goto out;

		/* Update written bytes count */
		written += cnt;
		/* Next write is aligned */
		col = 0;
		/* Increment page address */
		page++;
	}

out:
	if (ret == -EIO)
		this->cmdfunc(mtd, NAND_CMD_RESET, -1, -1);

	/* De-select the NAND device */
	nand_deselect ();

	/* Wake up anyone waiting on the device */
	spin_lock_bh (&this->chip_lock);
	this->state = FL_READY;
	wake_up (&this->wq);
	spin_unlock_bh (&this->chip_lock);

	*retlen = written;
	//printk("nand_write_ecc -> %d\n", ret);
	return ret;
}

/*
 * NAND write out-of-band
 */
static int nand_write_oob (struct mtd_info *mtd, loff_t to, size_t len, size_t * retlen, const u_char * buf)
{
	int i, column, page, ret = 0;
	struct nand_chip *this = mtd->priv;
	unsigned char data[DMC_NAND_LINE_SIZE_LONG];

	DEBUG (MTD_DEBUG_LEVEL3, "nand_write_oob: to = 0x%08x, len = %i\n", (unsigned int) to, (int) len);
	//printk("nand_write_oob: to = 0x%08x, len = %i\n", (unsigned int) to, (int) len);

	/* Shift to get page */
	page = ((int) to) >> this->page_shift;

	/* Mask to get column */
	column = to & 0x0f;

	/* Initialize return length value */
	*retlen = 0;

	/* Do not allow write past end of page */
	if ((column + len) > mtd->oobsize) {
		DEBUG (MTD_DEBUG_LEVEL0, "nand_write_oob: Attempt to write past end of page\n");
		//printk("nand_write_oob: Attempt read beyond end\n");
		return -EINVAL;
	}

	/* Grab the lock and see if the device is available */
	nand_get_chip (this, mtd, FL_WRITING, NULL);

	/* Select the NAND device */
	nand_select ();

	/* Check the WP bit */
	//this->cmdfunc (mtd, NAND_CMD_STATUS, -1, -1);
	//if (!(readb (this->IO_ADDR_R) & 0x80)) {
	//	DEBUG (MTD_DEBUG_LEVEL0, "nand_write_oob: Device is write protected!!!\n");
	//	ret = -EIO;
	//	goto out;
	//}

	/* Invalidate cache, if we write to this page */
	/* Write out desired data */
	nandReadPage(page, &data[0], &data[DMC_NAND_LINE_SIZE]);
	for (i = 0; i < len; i++)
		data[DMC_NAND_LINE_SIZE + column + i] = buf[i];
	nandWritePage(page, &data[0], &data[DMC_NAND_LINE_SIZE]);

	/* Return happy */
	*retlen = len;

#ifdef CONFIG_MTD_NAND_VERIFY_WRITE
#endif

	if (ret == -EIO)
		this->cmdfunc(mtd, NAND_CMD_RESET, -1, -1);

	/* De-select the NAND device */
	nand_deselect ();

	/* Wake up anyone waiting on the device */
	spin_lock_bh (&this->chip_lock);
	this->state = FL_READY;
	wake_up (&this->wq);
	spin_unlock_bh (&this->chip_lock);

	//printk("nand_write_oob -> %d\n", ret);
	return ret;
}


/*
 * NAND write with iovec
 */
static int nand_writev (struct mtd_info *mtd, const struct iovec *vecs, unsigned long count, 
		loff_t to, size_t * retlen)
{
	return (nand_writev_ecc (mtd, vecs, count, to, retlen, NULL, NULL));
}

static int nand_writev_ecc (struct mtd_info *mtd, const struct iovec *vecs, unsigned long count, 
		loff_t to, size_t * retlen, u_char *eccbuf, struct nand_oobinfo *oobsel)
{
	int i, page, col, cnt, len, total_len, ret = 0, written = 0;
	struct nand_chip *this = mtd->priv;

	/* Calculate total length of data */
	total_len = 0;
	for (i = 0; i < count; i++)
		total_len += (int) vecs[i].iov_len;

	DEBUG (MTD_DEBUG_LEVEL3,
	       "nand_writev: to = 0x%08x, len = %i, count = %ld\n", (unsigned int) to, (unsigned int) total_len, count);
	//printk("nand_writev_ecc: to = 0x%08x, len = %i\n", (unsigned int) to, (int) len);

	/* Do not allow write past end of page */
	if ((to + total_len) > mtd->size) {
		DEBUG (MTD_DEBUG_LEVEL0, "nand_writev: Attempted write past end of device\n");
		//printk("nand_writev_ecc: Attempt write past end\n");
		return -EINVAL;
	}

	/* Shift to get page */
	page = ((int) to) >> this->page_shift;

	/* Get the starting column */
	col = to & (mtd->oobblock - 1);

	/* Grab the lock and see if the device is available */
	nand_get_chip (this, mtd, FL_WRITING, NULL);

	/* Select the NAND device */
	nand_select ();

	/* Check the WP bit */
	//this->cmdfunc (mtd, NAND_CMD_STATUS, -1, -1);
	//if (!(readb (this->IO_ADDR_R) & 0x80)) {
	//	DEBUG (MTD_DEBUG_LEVEL0, "nand_writev: Device is write protected!!!\n");
	//	ret = -EIO;
	//	goto out;
	//}

	/* Loop until all iovecs' data has been written */
	cnt = col;
	len = 0;
	
	while (count) {
		/* 
		 *  Check, if we write from offset 0 and if the tuple
		 *  gives us not enough data for a full page write. Then we
		 *  can use the iov direct, else we have to copy into
		 *  data_buf.		
		 */
		if (!cnt && (vecs->iov_len - len) >= mtd->oobblock) {
			cnt = mtd->oobblock;
			this->data_poi = (u_char *) vecs->iov_base;
			this->data_poi += len;
			len += mtd->oobblock; 
			/* Check, if we have to switch to the next tuple */
			if (len >= (int) vecs->iov_len) {
				vecs++;
				len = 0;
				count--;
			}
		} else {

			/*
			 * Read data out of each tuple until we have a full page
			 * to write or we've read all the tuples.
		 	*/
			while ((cnt < mtd->oobblock) && count) {
				if (vecs->iov_base != NULL && vecs->iov_len) {
					this->data_buf[cnt++] = ((u_char *) vecs->iov_base)[len++];
				}
				/* Check, if we have to switch to the next tuple */
				if (len >= (int) vecs->iov_len) {
					vecs++;
					len = 0;
					count--;
				}
			}	
			this->data_poi = this->data_buf;	
		}
		
		/* We use the same function for write and writev !) */
		ret = nand_write_page (mtd, this, page, col, cnt, NULL, oobsel);
		if (ret)
			goto out;

		/* Update written bytes count */
		written += (cnt - col);

		/* Reset written byte counter and column */
		col = cnt = 0;

		/* Increment page address */
		page++;
	}

out:
	if (ret == -EIO)
		this->cmdfunc(mtd, NAND_CMD_RESET, -1, -1);

	/* De-select the NAND device */
	nand_deselect ();

	/* Wake up anyone waiting on the device */
	spin_lock_bh (&this->chip_lock);
	this->state = FL_READY;
	wake_up (&this->wq);
	spin_unlock_bh (&this->chip_lock);

	*retlen = written;
	//printk("nand_writev_ecc -> %d\n", ret);
	return ret;
}

/*
 * return true if hamming distance between a and 0xff is <= 1.
 *
 * [NOTE]
 *	assume that higher 3 bytes are zero
 */
static inline int nand_hamming_distance(unsigned int a)
{
	return (a == 0xff);
#if 0
#ifdef __ARM_ARCH_5TE__
	unsigned int n;
	a ^= 0xff;
	asm (
		"clz	%0, %1\n"
		: "=r" (n)
		: "r" (a)
	);
	return (a << (n + 1)) == 0;
#else
	return (a == 0xff ||
		a == 0x7f || a == 0xbf || a == 0xdf || a == 0xef ||
		a == 0xf7 || a == 0xfb || a == 0xfd || a == 0xfe);
#endif
#endif
}

/*
 * NAND erase a block
 */
static int nand_erase (struct mtd_info *mtd, struct erase_info *instr)
{
	int page, len, status, pages_per_block, ret;
	struct nand_chip *this = mtd->priv;
	DECLARE_WAITQUEUE (wait, current);

	DEBUG (MTD_DEBUG_LEVEL3,
	       "nand_erase: start = 0x%08x, len = %i\n", (unsigned int) instr->addr, (unsigned int) instr->len);
	//printk("nand_erase: start = 0x%08x, len = %i\n", (unsigned int) instr->addr, (unsigned int) instr->len);

	/* Start address must align on block boundary */
	if (instr->addr & (mtd->erasesize - 1)) {
		DEBUG (MTD_DEBUG_LEVEL0, "nand_erase: Unaligned address\n");
		printk("nand_erase: Unaligned address\n");
		return -EINVAL;
	}

	/* Length must align on block boundary */
	if (instr->len & (mtd->erasesize - 1)) {
		DEBUG (MTD_DEBUG_LEVEL0, "nand_erase: Length not block aligned\n");
		printk("nand_erase: Length not block aligned\n");
		return -EINVAL;
	}

	/* Do not allow erase past end of device */
	if ((instr->len + instr->addr) > mtd->size) {
		DEBUG (MTD_DEBUG_LEVEL0, "nand_erase: Erase past end of device\n");
		printk("nand_erase: Erase past end of device\n");
		return -EINVAL;
	}

	/* Grab the lock and see if the device is available */
	nand_get_chip (this, mtd, FL_ERASING, NULL);

	/* Shift to get first page */
	page = (int) (instr->addr >> this->page_shift);

	/* Calculate pages in each block */
	pages_per_block = mtd->erasesize / mtd->oobblock;

	/* Select the NAND device */
	nand_select ();

	/* Check the WP bit */
	//this->cmdfunc (mtd, NAND_CMD_STATUS, -1, -1);
	//if (!(readb (this->IO_ADDR_R) & 0x80)) {
	//	DEBUG (MTD_DEBUG_LEVEL0, "nand_erase: Device is write protected!!!\n");
	//	instr->state = MTD_ERASE_FAILED;
	//	goto erase_exit;
	//}

	/* Loop through the pages */
	len = instr->len;

	instr->state = MTD_ERASING;

	while (len) {
		/* Check if we have a bad block, we do not erase bad blocks ! */
#ifdef CONFIG_MTD_NAND_ERASE_BY_FORCE
		if (instr->by_force)
			goto skip_badblock_check;
#endif

		
		this->cmdfunc (mtd, NAND_CMD_READOOB, NAND_SMALL_BADBLOCK_POS, page);
		if (! nand_hamming_distance(nandCheckPage(page))) {
			printk ("%s: attempt to erase a bad block at addr 0x%08x\n", __func__, page << this->page_shift);
			instr->state = MTD_ERASE_FAILED;
			goto erase_exit;
		}
#ifdef CONFIG_MTD_NAND_POST_BADBLOCK
		this->cmdfunc (mtd, NAND_CMD_READOOB, NAND_POSTBADBLOCK_POS, page);
		if (! nand_hamming_distance(readb (this->IO_ADDR_R))) {
			printk ("%s: attempt to erase a post bad block at addr 0x%08x\n", __func__, page << this->page_shift);
			instr->state = MTD_ERASE_FAILED;
			goto erase_exit;
		}
#endif

#ifdef CONFIG_MTD_NAND_ERASE_BY_FORCE
	  skip_badblock_check:
#endif

		/* Send commands to erase a page */
		nandErasePage(page);

		spin_unlock_bh (&this->chip_lock);
		status = 0; //this->waitfunc (mtd, this, FL_ERASING);

		/* Get spinlock, in case we exit */
		spin_lock_bh (&this->chip_lock);
		/* See if block erase succeeded */
		if (status & 0x01) {
			DEBUG (MTD_DEBUG_LEVEL0, "nand_erase: " "Failed erase, page 0x%08x\n", page);
			printk("nand_erase: Failed erase page %x\n", page);
			instr->state = MTD_ERASE_FAILED;
			goto erase_exit;
		}
		
		/* Check, if we were interupted */
		if (this->state == FL_ERASING) {
			/* Invalidate cache, if last_page is inside erase-block */
			/* Increment page address and decrement length */
			len -= mtd->erasesize;
			page += pages_per_block;
		}
		/* Release the spin lock */
		spin_unlock_bh (&this->chip_lock);
erase_retry:
		spin_lock_bh (&this->chip_lock);
		/* Check the state and sleep if it changed */
		if (this->state == FL_ERASING || this->state == FL_READY) {
			/* Select the NAND device again, if we were interrupted */
			this->state = FL_ERASING;
			nand_select ();
			continue;
		} else {
			set_current_state (TASK_UNINTERRUPTIBLE);
			add_wait_queue (&this->wq, &wait);
			spin_unlock_bh (&this->chip_lock);
			schedule ();
			remove_wait_queue (&this->wq, &wait);
			goto erase_retry;
		}
	}
	instr->state = MTD_ERASE_DONE;

erase_exit:
	ret = instr->state == MTD_ERASE_DONE ? 0 : -EIO;
	if (ret == -EIO)
		this->cmdfunc(mtd, NAND_CMD_RESET, -1, -1);

	/* De-select the NAND device */
	nand_deselect ();
	spin_unlock_bh (&this->chip_lock);

	/* Do call back function */
	if (!ret && instr->callback)
		instr->callback (instr);

	/* The device is ready */
	spin_lock_bh (&this->chip_lock);
	this->state = FL_READY;
	wake_up (&this->wq);
	spin_unlock_bh (&this->chip_lock);

	/* Return more or less happy */
	//printk("nand_erase -> %d\n", ret);
	return ret;
}

/*
 * NAND sync
 */
static void nand_sync (struct mtd_info *mtd)
{
	struct nand_chip *this = mtd->priv;
	DECLARE_WAITQUEUE (wait, current);

	DEBUG (MTD_DEBUG_LEVEL3, "nand_sync: called\n");
	//printk("nand_sync\n");

retry:
	/* Grab the spinlock */
	spin_lock_bh (&this->chip_lock);

	/* See what's going on */
	switch (this->state) {
	case FL_READY:
	case FL_SYNCING:
		this->state = FL_SYNCING;
		spin_unlock_bh (&this->chip_lock);
		break;

	default:
		/* Not an idle state */
		add_wait_queue (&this->wq, &wait);
		spin_unlock_bh (&this->chip_lock);
		schedule ();

		remove_wait_queue (&this->wq, &wait);
		goto retry;
	}

	/* Lock the device */
	spin_lock_bh (&this->chip_lock);

	/* Set the device to be ready again */
	if (this->state == FL_SYNCING) {
		this->state = FL_READY;
		wake_up (&this->wq);
	}

	/* Unlock the device */
	spin_unlock_bh (&this->chip_lock);
}

int device_ioctl(struct inode *inode, struct file *file, unsigned int ioctl_num,
	unsigned long ioctl_param)
{
	int i;
	unsigned char data[DMC_NAND_LINE_SIZE_LONG];

	switch (ioctl_num) {
	case 1:
		printk("ReadTime:  %ld ms  Count: %ld\n", g_ReadTime, g_NumReads);
		printk("WriteTime: %ld ms  Count: %ld\n", g_WriteTime, g_NumWrites);
		printk("EraseTime: %ld ms  Count: %ld\n", g_EraseTime, g_NumErases);
		return 0;
	case 2:	
		g_ReadTime = g_WriteTime = g_EraseTime = 0;
		g_NumReads = g_NumWrites = g_NumErases = 0;
		return 0;
	case 3:
		for (i = 0; i < 5469; i++)
			nandReadPage(i+256, &data[0], &data[DMC_NAND_LINE_SIZE]);
		return 0;
	case 4:
		if(g_UseCache){
			g_UseCache = 0;
			nandCacheFree();
		}
		return 0;
	default:
		return -1;
	}
}

int device_open(struct inode *inode,struct file *file)
{
	return 0;
}

int device_release(struct inode *inode, struct file *file)
{
	return 0;
}

struct file_operations Fops = {
	NULL,                // struct module *owner;
	NULL, /* seek */		// loff_t (*llseek) (struct file *, loff_t, int);
	NULL,				      // ssize_t (*read) (struct file *, char *, size_t, loff_t *);
	NULL,				      // ssize_t (*write) (struct file *, const char *, size_t, loff_t *);
	NULL, /* readdir */  // int (*readdir) (struct file *, void *, filldir_t);
	NULL, /* select */   // unsigned int (*poll) (struct file *, struct poll_table_struct *);
	device_ioctl,	      // int (*ioctl) (struct inode *, struct file *, unsigned int, unsigned long);
	NULL, /* mmap */     // int (*mmap) (struct file *, struct vm_area_struct *);
	device_open,	      // int (*open) (struct inode *, struct file *);
	NULL, /* flush */    // int (*flush) (struct file *);
	device_release,	   // int (*release) (struct inode *, struct file *);
	NULL, 			      // int (*fsync) (struct file *, struct dentry *, int datasync);
	NULL, 			      // int (*fasync) (int, struct file *, int);
	NULL, 			      // int (*lock) (struct file *, int, struct file_lock *);
	NULL, 			      // ssize_t (*readv) (struct file *, const struct iovec *, unsigned long, loff_t *);
	NULL, 		      	// ssize_t (*writev) (struct file *, const struct iovec *, unsigned long, loff_t *);
	NULL, 		     		// ssize_t (*sendpage) (struct file *, struct page *, int, size_t, loff_t *, int);
	NULL 	 	 		      // unsigned long (*get_unmapped_area)(struct file *, unsigned long, unsigned long, unsigned long, unsigned long);
};

static int nand_procinfo(char* buf, char** start, off_t offset,
			 int count, int* eof, void* data)
{
    int len = 0;

    len += sprintf(buf + len, "DMC NAND Status:\n");
    len += sprintf(buf + len, "Cache pages:  %ld\n", g_NAND_CachePages);
    len += sprintf(buf + len, "Cache hits:   %ld\n", g_NAND_CacheHits);
    len += sprintf(buf + len, "Cache misses: %ld\n", g_NAND_CacheMisses);

    ehci_show_registers();
    uhci_show_registers();

    return len;
}

/*
 * NAND Flash driver Initialize/Exit
 */	
int __init nand_init_module(void)
{
	struct mtd_info  *mtd;
	struct nand_chip *this = 0;
	struct pci_dev *FPGA_dev;
	unsigned long FPGA_io_base;
	unsigned long FPGA_io_end;

	printk("dmc_nand: init_module.\n");

	nand_part_info[0].offset = nand_start;
	nand_part_info[0].size   = nand_size - nand_start;

#ifdef NAND_USE_TIMER
	FPGA_dev = pci_find_device(DMC_FPGA_VENDOR_ID, DMC_FPGA_DEVICE_ID, NULL);
	if(! FPGA_dev ) {
		printk("dmc_nand: FPGA Device not found !\n");
	}
	else {
		FPGA_io_base = pci_resource_start(FPGA_dev, 0);
		FPGA_io_end = pci_resource_end(FPGA_dev, 0);

		FPGA_io_base_address = (unsigned long) ioremap_nocache(TX7901_PCI_IO_LOWER + FPGA_io_base, (FPGA_io_end - FPGA_io_base));
		printk("dmc_nand: FPGA device found at 0x%08lx\n", FPGA_io_base_address);

		FPGA_CLOCK = (vu32*) FPGA_io_base_address + (FPGA_CLOCK_OFFSET / sizeof(vu32));
	}
#endif

	/*  Allocate memory for MTD device structure and private data  */
	mtd = kmalloc(sizeof(struct mtd_info), GFP_KERNEL);
	if( !mtd ) {
		printk (KERN_ERR "Unable to allocate DMC NAND MTD device structure.\n");
		return -ENOMEM;
	}
	this = kmalloc(sizeof (struct nand_chip), GFP_KERNEL);
	if( !this ) {
		printk (KERN_ERR "Unable to allocate DMC NAND MTD device structure.\n");
		return -ENOMEM;
	}
	/*  Initialize structures  */
	memset((char *)mtd, 0, sizeof(struct mtd_info));
	memset((char *)this, 0, sizeof(struct nand_chip));

	/*  Set nand_chip  */
	this->chip_delay      = 15;
	this->hwcontrol	= NULL;
	this->cmdfunc	      = nand_command;
	this->waitfunc	= nand_wait;
	this->dev_ready	= nand_dev_ready;
	this->page_shift      = 9;
	this->state           = FL_READY;
	this->autooob = &nand_oob_16;
	init_waitqueue_head (&this->wq);
	spin_lock_init (&this->chip_lock);

	/* set mtd info */
	mtd->priv       = this;
	mtd->type       = MTD_NANDFLASH;
	mtd->flags      = MTD_CAP_NANDFLASH | MTD_ECC | MTD_WRITEABLE;
	mtd->ecctype    = MTD_ECC_SW;
	mtd->erase      = nand_erase;
	mtd->point      = NULL;
	mtd->unpoint    = NULL;
	mtd->read       = nand_read;
	mtd->write      = nand_write;
	mtd->read_ecc   = nand_read_ecc;
	mtd->write_ecc  = nand_write_ecc;
	mtd->read_oob   = nand_read_oob;
	mtd->write_oob  = nand_write_oob;
	mtd->readv      = NULL;
	mtd->writev     = nand_writev;
	mtd->writev_ecc = nand_writev_ecc;
	mtd->sync       = nand_sync;
	mtd->lock       = NULL;
	mtd->unlock     = NULL;
	mtd->suspend    = NULL;
	mtd->resume     = NULL;
	mtd->block_isbad = nand_block_isbad;
	mtd->block_markbad = nand_block_markbad;

	mtd->name       = "dmc_nand";
	mtd->erasesize  = DMC_NAND_BLOCK_SIZE;
	mtd->size       = DMC_NAND_SIZE * 2;
	mtd->eccsize    = 256;
	mtd->oobblock   = DMC_NAND_LINE_SIZE;
	mtd->oobsize    = DMC_NAND_LINE_SIZE_LONG - DMC_NAND_LINE_SIZE;
	mtd->oobavail   = mtd->oobsize - (this->autooob->eccbytes + 1);
	memcpy(&mtd->oobinfo, this->autooob, sizeof(mtd->oobinfo));

	/*  Allocate memory for internal data buffer  */
	this->data_buf = kmalloc (sizeof(u_char) * (mtd->oobblock + mtd->oobsize), GFP_KERNEL);
	if( !this->data_buf ) {
		printk (KERN_ERR "Unable to allocate NAND data buffer.\n");
		kfree(this);
		kfree(mtd);
		return -ENOMEM;
	}

	g_UseCache = 1;
	if(g_UseCache){
		nandCacheInit();
	}
	/* Register the partitions */
	add_mtd_partitions(mtd, nand_part_info, NAND_NPARTS);

	/* Register the char device */
	register_chrdev(191, "dmc_nand", &Fops);
	create_proc_read_entry("driver/dmc_nand", 0, NULL, nand_procinfo, NULL);
  
	printk("%s initialized.\n", DEV_NAME);

	return 0;	/*  Return happy  */
}
module_init(nand_init_module);


MODULE_LICENSE ("GPL");
MODULE_AUTHOR ("Steven J. Hill <sjhill@cotw.com>, Thomas Gleixner <tglx@linutronix.de>");
MODULE_DESCRIPTION ("Generic NAND flash driver code");

/*
 * Local variables:
 *   c-basic-offset: 8
 * End:
 */
